# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import io
import json
import os
import textwrap
from datetime import datetime
from unittest.mock import Mock, PropertyMock, patch
from urllib.error import URLError

import networkx as nx
import pytest
from botocore.exceptions import ClientError
from common_test_utils import (
    DM1_ARN,
    DWAVE_ARN,
    IONQ_ARN,
    OQC_ARN,
    RIGETTI_ARN,
    RIGETTI_REGION,
    SV1_ARN,
    TN1_ARN,
    run_and_assert,
    run_batch_and_assert,
)
from jsonschema import validate

from braket.aws import AwsDevice, AwsDeviceType, AwsQuantumTask
from braket.aws.queue_information import QueueDepthInfo, QueueType
from braket.circuits import Circuit, FreeParameter, Gate, Noise, QubitSet
from braket.circuits.gate_calibrations import GateCalibrations
from braket.circuits.noise_model import GateCriteria, NoiseModel
from braket.device_schema.device_execution_window import DeviceExecutionWindow
from braket.device_schema.dwave import DwaveDeviceCapabilities
from braket.device_schema.rigetti import RigettiDeviceCapabilities
from braket.device_schema.simulators import GateModelSimulatorDeviceCapabilities
from braket.ir.openqasm import Program as OpenQasmProgram
from braket.pulse import DragGaussianWaveform, Frame, Port, PulseSequence

MOCK_GATE_MODEL_QPU_CAPABILITIES_JSON_1 = {
    "braketSchemaHeader": {
        "name": "braket.device_schema.rigetti.rigetti_device_capabilities",
        "version": "1",
    },
    "service": {
        "executionWindows": [
            {
                "executionDay": "Everyday",
                "windowStartHour": "11:00",
                "windowEndHour": "12:00",
            }
        ],
        "shotsRange": [1, 10],
    },
    "action": {
        "braket.ir.jaqcd.program": {
            "actionType": "braket.ir.jaqcd.program",
            "version": ["1"],
            "supportedOperations": ["H"],
        }
    },
    "paradigm": {
        "qubitCount": 30,
        "nativeGateSet": ["ccnot", "cy"],
        "connectivity": {"fullyConnected": False, "connectivityGraph": {"1": ["2", "3"]}},
    },
    "deviceParameters": {},
}

MOCK_GATE_MODEL_QPU_CAPABILITIES_1 = RigettiDeviceCapabilities.parse_obj(
    MOCK_GATE_MODEL_QPU_CAPABILITIES_JSON_1
)

MOCK_gate_calibrations_JSON = {
    "gates": {
        "0": {
            "cphaseshift": [
                {
                    "name": "cphaseshift",
                    "qubits": ["0", "1"],
                    "arguments": ["-1.5707963267948966"],
                    "calibrations": [
                        {
                            "name": "barrier",
                            "arguments": [{"name": "qubit", "value": "0", "type": "string"}],
                        },
                        {
                            "name": "play",
                            "arguments": [
                                {"name": "frame", "value": "q0_q1_cphase_frame", "type": "frame"},
                                {
                                    "name": "waveform",
                                    "value": "wf_drag_gaussian_0",
                                    "type": "waveform",
                                },
                            ],
                        },
                        {
                            "name": "barrier",
                            "arguments": [
                                {"name": "frame", "value": "q0_q1_cphase_frame", "type": "frame"}
                            ],
                        },
                        {
                            "name": "barrier",
                            "arguments": None,
                        },
                        {
                            "name": "delay",
                            "arguments": [
                                {"name": "duration", "value": 3e-07, "type": "float"},
                                {"name": "qubit", "value": "0", "type": "string"},
                                {"name": "qubit", "value": "1", "type": "string"},
                            ],
                        },
                        {
                            "name": "delay",
                            "arguments": [
                                {"name": "frame", "value": "q0_q1_cphase_frame", "type": "frame"},
                                {"name": "duration", "value": 3e-07, "type": "float"},
                            ],
                        },
                        {
                            "name": "shift_phase",
                            "arguments": [
                                {"name": "frame", "value": "q0_q1_cphase_frame", "type": "frame"},
                                {"name": "phase", "value": 3e-07, "type": "float"},
                            ],
                        },
                        {
                            "name": "shift_frequency",
                            "arguments": [
                                {"name": "frequency", "value": "theta", "type": "expr"},
                                {"name": "frame", "value": "q0_q1_cphase_frame", "type": "frame"},
                                {"name": "extra", "value": "q0_q1_cphase_frame", "type": "string"},
                            ],
                        },
                    ],
                }
            ],
            "rx": [
                {
                    "gateName": "rx",
                    "gateId": "rz_1",
                    "qubits": "0",
                    "arguments": ["theta"],
                    "calibrations": [
                        {"name": "barrier", "arguments": None},
                    ],
                }
            ],
        },
        "0_1": {
            "cz": [
                {
                    "gateName": "cz",
                    "gateId": "cz_0_1",
                    "qubits": ["1", "0"],
                    "arguments": [],
                    "calibrations": [
                        {"name": "barrier", "arguments": None},
                    ],
                }
            ],
            "rx_12": [],
        },
    },
    "waveforms": {
        "q0_q1_cz_CZ": {
            "waveformId": "q0_q1_cz_CZ",
            "amplitudes": [[0.0, 0.0], [0.0, 0.0]],
        },
        "wf_drag_gaussian_0": {
            "waveformId": "wf_drag_gaussian_0",
            "name": "drag_gaussian",
            "arguments": [
                {"name": "length", "value": 6.000000000000001e-8, "type": "float"},
                {"name": "sigma", "value": 6.369913502160144e-9, "type": "float"},
                {"name": "amplitude", "value": -0.4549282253548838, "type": "float"},
                {"name": "beta", "value": 7.494904522022295e-10, "type": "float"},
            ],
        },
        "wf_gaussian_0": {
            "waveformId": "wf_gaussian_0",
            "name": "gaussian",
            "arguments": [
                {"name": "length", "value": 6.000000000000001e-8, "type": "float"},
                {"name": "sigma", "value": 6.369913502160144e-9, "type": "float"},
                {"name": "amplitude", "value": -0.4549282253548838, "type": "float"},
            ],
        },
        "wf_constant": {
            "waveformId": "wf_constant",
            "name": "constant",
            "arguments": [
                {"name": "length", "value": 2, "type": "float"},
                {"name": "iq", "value": 0.23, "type": "complex"},
            ],
        },
    },
}


def test_mock_rigetti_schema_1():
    validate(MOCK_GATE_MODEL_QPU_CAPABILITIES_JSON_1, RigettiDeviceCapabilities.schema())


MOCK_GATE_MODEL_QPU_1 = {
    "deviceName": "Aspen-10",
    "deviceType": "QPU",
    "providerName": "Rigetti",
    "deviceStatus": "OFFLINE",
    "deviceCapabilities": MOCK_GATE_MODEL_QPU_CAPABILITIES_1.json(),
    "deviceQueueInfo": [
        {"queue": "QUANTUM_TASKS_QUEUE", "queueSize": "19", "queuePriority": "Normal"},
        {"queue": "QUANTUM_TASKS_QUEUE", "queueSize": "3", "queuePriority": "Priority"},
        {"queue": "JOBS_QUEUE", "queueSize": "0 (3 prioritized job(s) running)"},
    ],
}

MOCK_GATE_MODEL_QPU_CAPABILITIES_JSON_2 = {
    "braketSchemaHeader": {
        "name": "braket.device_schema.rigetti.rigetti_device_capabilities",
        "version": "1",
    },
    "service": {
        "executionWindows": [
            {
                "executionDay": "Everyday",
                "windowStartHour": "11:00",
                "windowEndHour": "12:00",
            }
        ],
        "shotsRange": [1, 10],
    },
    "action": {
        "braket.ir.jaqcd.program": {
            "actionType": "braket.ir.jaqcd.program",
            "version": ["1"],
            "supportedOperations": ["H"],
        }
    },
    "paradigm": {
        "qubitCount": 30,
        "nativeGateSet": ["ccnot", "cy"],
        "connectivity": {"fullyConnected": True, "connectivityGraph": {}},
    },
    "deviceParameters": {},
}

MOCK_GATE_MODEL_QPU_CAPABILITIES_2 = RigettiDeviceCapabilities.parse_obj(
    MOCK_GATE_MODEL_QPU_CAPABILITIES_JSON_2
)


def test_mock_rigetti_schema_2():
    validate(MOCK_GATE_MODEL_QPU_CAPABILITIES_JSON_2, RigettiDeviceCapabilities.schema())


MOCK_GATE_MODEL_QPU_2 = {
    "deviceName": "Blah",
    "deviceType": "QPU",
    "providerName": "blahhhh",
    "deviceStatus": "OFFLINE",
    "deviceCapabilities": MOCK_GATE_MODEL_QPU_CAPABILITIES_2.json(),
}

MOCK_GATE_MODEL_QPU_3 = {
    "deviceName": "Lucy",
    "deviceType": "QPU",
    "providerName": "OQC",
    "deviceStatus": "OFFLINE",
    "deviceCapabilities": MOCK_GATE_MODEL_QPU_CAPABILITIES_1.json(),
}

MOCK_DWAVE_QPU_CAPABILITIES_JSON = {
    "braketSchemaHeader": {
        "name": "braket.device_schema.dwave.dwave_device_capabilities",
        "version": "1",
    },
    "provider": {
        "annealingOffsetStep": 1.45,
        "annealingOffsetStepPhi0": 1.45,
        "annealingOffsetRanges": [[1.45, 1.45], [1.45, 1.45]],
        "annealingDurationRange": [1, 2, 3],
        "couplers": [[1, 2], [2, 3]],
        "defaultAnnealingDuration": 1,
        "defaultProgrammingThermalizationDuration": 1,
        "defaultReadoutThermalizationDuration": 1,
        "extendedJRange": [1, 2, 3],
        "hGainScheduleRange": [1, 2, 3],
        "hRange": [1, 2, 3],
        "jRange": [1, 2, 3],
        "maximumAnnealingSchedulePoints": 1,
        "maximumHGainSchedulePoints": 1,
        "perQubitCouplingRange": [1, 2, 3],
        "programmingThermalizationDurationRange": [1, 2, 3],
        "qubits": [1, 2, 3],
        "qubitCount": 1,
        "quotaConversionRate": 1,
        "readoutThermalizationDurationRange": [1, 2, 3],
        "taskRunDurationRange": [1, 2, 3],
        "topology": {},
    },
    "service": {
        "executionWindows": [
            {"executionDay": "Everyday", "windowStartHour": "11:00", "windowEndHour": "12:00"}
        ],
        "shotsRange": [1, 10],
    },
    "action": {
        "braket.ir.annealing.problem": {
            "actionType": "braket.ir.annealing.problem",
            "version": ["1"],
        }
    },
    "deviceParameters": {},
}

MOCK_DWAVE_QPU_CAPABILITIES = DwaveDeviceCapabilities.parse_obj(MOCK_DWAVE_QPU_CAPABILITIES_JSON)


def test_d_wave_schema():
    validate(MOCK_DWAVE_QPU_CAPABILITIES_JSON, DwaveDeviceCapabilities.schema())


MOCK_DWAVE_QPU = {
    "deviceName": "Advantage_system1.1",
    "deviceType": "QPU",
    "providerName": "provider1",
    "deviceStatus": "ONLINE",
    "deviceCapabilities": MOCK_DWAVE_QPU_CAPABILITIES.json(),
}

MOCK_GATE_MODEL_SIMULATOR_CAPABILITIES_JSON = {
    "braketSchemaHeader": {
        "name": "braket.device_schema.simulators.gate_model_simulator_device_capabilities",
        "version": "1",
    },
    "service": {
        "executionWindows": [
            {
                "executionDay": "Everyday",
                "windowStartHour": "11:00",
                "windowEndHour": "12:00",
            }
        ],
        "shotsRange": [1, 10],
    },
    "action": {
        "braket.ir.jaqcd.program": {
            "actionType": "braket.ir.jaqcd.program",
            "version": ["1"],
            "supportedOperations": ["H"],
        },
    },
    "paradigm": {"qubitCount": 30},
    "deviceParameters": {},
}

MOCK_GATE_MODEL_NOISE_SIMULATOR_CAPABILITIES_JSON = {
    "braketSchemaHeader": {
        "name": "braket.device_schema.simulators.gate_model_simulator_device_capabilities",
        "version": "1",
    },
    "service": {
        "executionWindows": [
            {
                "executionDay": "Everyday",
                "windowStartHour": "11:00",
                "windowEndHour": "12:00",
            }
        ],
        "shotsRange": [1, 10],
    },
    "action": {
        "braket.ir.openqasm.program": {
            "actionType": "braket.ir.openqasm.program",
            "version": ["1"],
            "supportedOperations": ["rx", "ry", "h", "cy", "cnot", "unitary"],
            "supportedResultTypes": [
                {
                    "name": "StateVector",
                    "observables": ["x", "y", "z"],
                    "minShots": 0,
                    "maxShots": 0,
                },
            ],
            "supportedPragmas": [
                "braket_noise_bit_flip",
                "braket_noise_depolarizing",
                "braket_noise_kraus",
                "braket_noise_pauli_channel",
                "braket_noise_generalized_amplitude_damping",
                "braket_noise_amplitude_damping",
                "braket_noise_phase_flip",
                "braket_noise_phase_damping",
                "braket_noise_two_qubit_dephasing",
                "braket_noise_two_qubit_depolarizing",
                "braket_unitary_matrix",
                "braket_result_type_sample",
                "braket_result_type_expectation",
                "braket_result_type_variance",
                "braket_result_type_probability",
                "braket_result_type_density_matrix",
            ],
        },
    },
    "paradigm": {"qubitCount": 30},
    "deviceParameters": {},
}

MOCK_GATE_MODEL_SIMULATOR_CAPABILITIES = GateModelSimulatorDeviceCapabilities.parse_obj(
    MOCK_GATE_MODEL_SIMULATOR_CAPABILITIES_JSON
)


def test_gate_model_sim_schema():
    validate(
        MOCK_GATE_MODEL_SIMULATOR_CAPABILITIES_JSON, GateModelSimulatorDeviceCapabilities.schema()
    )


MOCK_GATE_MODEL_NOISE_SIMULATOR_CAPABILITIES = GateModelSimulatorDeviceCapabilities.parse_obj(
    MOCK_GATE_MODEL_NOISE_SIMULATOR_CAPABILITIES_JSON
)


def test_gate_model_sim_schema():
    validate(
        MOCK_GATE_MODEL_NOISE_SIMULATOR_CAPABILITIES_JSON,
        GateModelSimulatorDeviceCapabilities.schema(),
    )


MOCK_GATE_MODEL_SIMULATOR = {
    "deviceName": "SV1",
    "deviceType": "SIMULATOR",
    "providerName": "provider1",
    "deviceStatus": "ONLINE",
    "deviceCapabilities": MOCK_GATE_MODEL_SIMULATOR_CAPABILITIES.json(),
}


MOCK_GATE_MODEL_NOISE_SIMULATOR = {
    "deviceName": "DM1",
    "deviceType": "SIMULATOR",
    "providerName": "provider1",
    "deviceStatus": "ONLINE",
    "deviceCapabilities": MOCK_GATE_MODEL_NOISE_SIMULATOR_CAPABILITIES.json(),
}


MOCK_DEFAULT_S3_DESTINATION_FOLDER = (
    "amazon-braket-us-test-1-00000000",
    "tasks",
)


@pytest.fixture(
    params=[
        "arn:aws:braket:us-west-1::device/quantum-simulator/amazon/sim",
        "arn:aws:braket:::device/quantum-simulator/amazon/sim",
    ]
)
def arn(request):
    return request.param


@pytest.fixture
def s3_destination_folder():
    return "bucket-foo", "key-bar"


@pytest.fixture
def bell_circuit():
    return Circuit().h(0)


@pytest.fixture
def openqasm_program():
    return OpenQasmProgram(source="OPENQASM 3.0; h $0;")


@pytest.fixture(params=["bell_circuit", "openqasm_program"])
def circuit(request):
    return request.getfixturevalue(request.param)


@pytest.fixture
def multiple_circuit_inputs():
    theta = FreeParameter("theta")
    beta = FreeParameter("beta")
    return Circuit().ry(angle=theta, target=0).rx(angle=beta, target=1)


@pytest.fixture()
def single_circuit_input():
    theta = FreeParameter("theta")
    return Circuit().ry(angle=theta, target=0)


@pytest.fixture
def aws_session():
    _boto_session = Mock()
    _boto_session.region_name = RIGETTI_REGION
    _boto_session.profile_name = "test-profile"

    creds = Mock()
    creds.method = "other"
    _boto_session.get_credentials.return_value = creds

    _aws_session = Mock()
    _aws_session.boto_session = _boto_session
    _aws_session._default_bucket = MOCK_DEFAULT_S3_DESTINATION_FOLDER[0]
    _aws_session.default_bucket.return_value = _aws_session._default_bucket
    _aws_session._custom_default_bucket = False
    _aws_session.account_id = "00000000"
    _aws_session.region = RIGETTI_REGION
    return _aws_session


@pytest.fixture
def device(aws_session):
    def _device(arn):
        aws_session.get_device.return_value = MOCK_GATE_MODEL_QPU_1
        aws_session.search_devices.return_value = [MOCK_GATE_MODEL_QPU_1]
        return AwsDevice(arn, aws_session)

    return _device


@pytest.mark.parametrize(
    "device_capabilities, get_device_data",
    [
        (MOCK_GATE_MODEL_SIMULATOR_CAPABILITIES, MOCK_GATE_MODEL_SIMULATOR),
        (MOCK_GATE_MODEL_QPU_CAPABILITIES_1, MOCK_GATE_MODEL_QPU_1),
        (MOCK_DWAVE_QPU_CAPABILITIES, MOCK_DWAVE_QPU),
    ],
)
def test_device_aws_session(device_capabilities, get_device_data, arn):
    mock_session = Mock()
    mock_session.get_device.return_value = get_device_data
    mock_session.region = RIGETTI_REGION
    device = AwsDevice(arn, mock_session)
    _assert_device_fields(device, device_capabilities, get_device_data)
    assert device.aws_session == mock_session


@patch("braket.aws.aws_device.AwsSession")
def test_device_simulator_no_aws_session(aws_session_init, aws_session):
    arn = SV1_ARN
    aws_session_init.return_value = aws_session
    aws_session.get_device.return_value = MOCK_GATE_MODEL_SIMULATOR
    device = AwsDevice(arn)
    _assert_device_fields(device, MOCK_GATE_MODEL_SIMULATOR_CAPABILITIES, MOCK_GATE_MODEL_SIMULATOR)
    aws_session.get_device.assert_called_with(arn)


@patch("braket.aws.aws_device.AwsSession.copy_session")
@patch("braket.aws.aws_device.AwsSession")
@pytest.mark.parametrize(
    "get_device_side_effect",
    [
        [MOCK_GATE_MODEL_QPU_1],
        [
            ClientError(
                {
                    "Error": {
                        "Code": "ResourceNotFoundException",
                    }
                },
                "getDevice",
            ),
            MOCK_GATE_MODEL_QPU_1,
        ],
    ],
)
def test_device_qpu_no_aws_session(
    aws_session_init, mock_copy_session, get_device_side_effect, aws_session
):
    arn = RIGETTI_ARN
    mock_session = Mock()
    mock_session.get_device.side_effect = get_device_side_effect
    aws_session.get_device.side_effect = ClientError(
        {
            "Error": {
                "Code": "ResourceNotFoundException",
            }
        },
        "getDevice",
    )
    aws_session_init.return_value = aws_session
    mock_copy_session.return_value = mock_session
    device = AwsDevice(arn)
    _assert_device_fields(device, MOCK_GATE_MODEL_QPU_CAPABILITIES_1, MOCK_GATE_MODEL_QPU_1)


@patch("braket.aws.aws_device.AwsSession.copy_session")
@patch("braket.aws.aws_device.AwsSession")
def test_regional_device_region_switch(aws_session_init, mock_copy_session, aws_session):
    device_region = "device-region"
    arn = f"arn:aws:braket:{device_region}::device/quantum-simulator/amazon/sim"
    aws_session_init.return_value = aws_session
    mock_session = Mock()
    mock_session.get_device.return_value = MOCK_GATE_MODEL_SIMULATOR
    mock_copy_session.return_value = mock_session
    device = AwsDevice(arn)
    aws_session.get_device.assert_not_called()
    mock_copy_session.assert_called_once()
    mock_copy_session.assert_called_with(aws_session, device_region)
    _assert_device_fields(device, MOCK_GATE_MODEL_SIMULATOR_CAPABILITIES, MOCK_GATE_MODEL_SIMULATOR)


@patch("braket.aws.aws_device.AwsSession")
@pytest.mark.parametrize(
    "get_device_side_effect, expected_exception",
    [
        (
            [
                ClientError(
                    {
                        "Error": {
                            "Code": "ResourceNotFoundException",
                        }
                    },
                    "getDevice",
                )
            ],
            ValueError,
        ),
        (
            [
                ClientError(
                    {
                        "Error": {
                            "Code": "ThrottlingException",
                        }
                    },
                    "getDevice",
                )
            ],
            ClientError,
        ),
    ],
)
def test_regional_device_raises_error(
    aws_session_init, get_device_side_effect, expected_exception, aws_session
):
    arn = "arn:aws:braket:us-west-1::device/quantum-simulator/amazon/sim"
    aws_session.get_device.side_effect = get_device_side_effect
    aws_session_init.return_value = aws_session
    with pytest.raises(expected_exception):
        AwsDevice(arn)
        aws_session.get_device.assert_called_once()


def test_device_refresh_metadata(arn):
    mock_session = Mock()
    mock_session.get_device.return_value = MOCK_GATE_MODEL_QPU_1
    mock_session.region = RIGETTI_REGION
    device = AwsDevice(arn, mock_session)
    _assert_device_fields(device, MOCK_GATE_MODEL_QPU_CAPABILITIES_1, MOCK_GATE_MODEL_QPU_1)

    mock_session.get_device.return_value = MOCK_GATE_MODEL_QPU_2
    device.refresh_metadata()
    _assert_device_fields(device, MOCK_GATE_MODEL_QPU_CAPABILITIES_2, MOCK_GATE_MODEL_QPU_2)


MOCK_PULSE_MODEL_QPU_PULSE_CAPABILITIES_JSON_1 = {
    "braketSchemaHeader": {
        "name": "braket.device_schema.pulse.pulse_device_action_properties",
        "version": "1",
    },
    "supportedQhpTemplateWaveforms": {},
    "supportedFunctions": {},
    "ports": {
        "q0_ff": {
            "portId": "q0_ff",
            "direction": "tx",
            "portType": "ff",
            "dt": 1e-09,
            "qubitMappings": None,
            "centerFrequencies": [375000000.0],
            "qhpSpecificProperties": None,
        }
    },
    "frames": {
        "q0_q1_cphase_frame": {
            "frameId": "q0_q1_cphase_frame",
            "portId": "q0_ff",
            "frequency": 4276236.85736918,
            "centerFrequency": 375000000.0,
            "phase": 1.0,
            "associatedGate": "cphase",
            "qubitMappings": [0, 1],
            "qhpSpecificProperties": None,
        }
    },
    "nativeGateCalibrationsRef": "file://hostname/foo/bar",
}

MOCK_PULSE_MODEL_QPU_PULSE_CAPABILITIES_JSON_2 = {
    "braketSchemaHeader": {
        "name": "braket.device_schema.pulse.pulse_device_action_properties",
        "version": "1",
    },
    "supportedQhpTemplateWaveforms": {},
    "supportedFunctions": {},
    "ports": {
        "q0_ff": {
            "portId": "q0_ff",
            "direction": "tx",
            "portType": "ff",
            "dt": 1e-09,
            "qubitMappings": None,
            "centerFrequencies": [375000000.0],
            "qhpSpecificProperties": None,
        }
    },
    "nativeGateCalibrationssRef": "file://hostname/foo/bar",
}


def get_pulse_model(capabilities_json):
    device_json = {
        "braketSchemaHeader": {
            "name": "braket.device_schema.rigetti.rigetti_device_capabilities",
            "version": "1",
        },
        "service": {
            "executionWindows": [
                {
                    "executionDay": "Everyday",
                    "windowStartHour": "11:00",
                    "windowEndHour": "12:00",
                }
            ],
            "shotsRange": [1, 10],
        },
        "provider": {
            "specs": {
                "1Q": {
                    "0": {
                        "fActiveReset": 0.9715,
                        "fRO": 0.951,
                        "f1QRB": 0.997339217568556,
                        "f1QRB_std_err": 0.00006690422818326937,
                        "f1Q_simultaneous_RB": 0.9949723201166536,
                        "f1Q_simultaneous_RB_std_err": 0.00021695233492231294,
                        "T1": 0.000010019627401991471,
                        "T2": 0.000018156447816365015,
                    }
                },
                "2Q": {
                    "0-1": {
                        "fCZ": 0.9586440436264603,
                        "fCZ_std_err": 0.007025921432645824,
                        "fCPHASE": 0.9287330972713645,
                        "fCPHASE_std_err": 0.009709406809550082,
                        "fXY": 0.9755179214520402,
                        "fXY_std_err": 0.0060234488782598536,
                    },
                },
            }
        },
        "action": {
            "braket.ir.jaqcd.program": {
                "actionType": "braket.ir.jaqcd.program",
                "version": ["1"],
                "supportedOperations": ["H"],
            }
        },
        "paradigm": {
            "qubitCount": 30,
            "nativeGateSet": ["ccnot", "cy"],
            "connectivity": {"fullyConnected": False, "connectivityGraph": {"1": ["2", "3"]}},
        },
        "deviceParameters": {},
        "pulse": capabilities_json,
    }
    device_obj = RigettiDeviceCapabilities.parse_obj(device_json)
    return {
        "deviceName": "M-2-Pulse",
        "deviceType": "QPU",
        "providerName": "Rigetti",
        "deviceStatus": "OFFLINE",
        "deviceCapabilities": device_obj.json(),
    }


@pytest.mark.parametrize(
    "pulse_device_capabilities",
    [
        MOCK_PULSE_MODEL_QPU_PULSE_CAPABILITIES_JSON_1,
        MOCK_PULSE_MODEL_QPU_PULSE_CAPABILITIES_JSON_2,
    ],
)
def test_device_pulse_metadata(pulse_device_capabilities):
    mock_session = Mock()
    mock_session.get_device.return_value = get_pulse_model(pulse_device_capabilities)
    mock_session.region = RIGETTI_REGION
    device = AwsDevice("arn:aws:braket:us-west-1::TestName", mock_session)
    assert device.ports == {"q0_ff": Port("q0_ff", 1e-9)}
    port = device.ports["q0_ff"]
    assert port.properties == pulse_device_capabilities["ports"]["q0_ff"]
    if "frames" in pulse_device_capabilities:
        assert device.frames == {
            "q0_q1_cphase_frame": Frame(
                "q0_q1_cphase_frame", Port("q0_ff", 1e-9), 4276236.85736918, 1.0
            )
        }
        frame = device.frames["q0_q1_cphase_frame"]
        assert frame.is_predefined is True
        assert frame.properties == pulse_device_capabilities["frames"]["q0_q1_cphase_frame"]
    else:
        assert device.frames == {}


def test_gate_calibration_refresh_no_url(arn):
    mock_session = Mock()
    mock_session.get_device.return_value = MOCK_GATE_MODEL_QPU_1
    mock_session.region = RIGETTI_REGION
    device = AwsDevice(arn, mock_session)

    assert device.refresh_gate_calibrations() == None


@patch("urllib.request.urlopen")
def test_device_gate_calibrations_exists(mock_url_request):
    # The data is accessed using a device manager so here data is prepped and passed for the return val.
    response_data_content = {
        "gates": {
            "0_1": {
                "cphaseshift": [
                    {
                        "name": "cphaseshift",
                        "qubits": ["0", "1"],
                        "arguments": ["-1.5707963267948966"],
                        "calibrations": [
                            {
                                "name": "play",
                                "arguments": [
                                    {
                                        "name": "waveform",
                                        "value": "wf_drag_gaussian_0",
                                        "type": "waveform",
                                    },
                                    {
                                        "name": "frame",
                                        "value": "q0_q1_cphase_frame",
                                        "type": "frame",
                                    },
                                ],
                            },
                        ],
                    }
                ],
                "rx_12": [{"name": "rx_12", "qubits": ["0"]}],
            },
        },
        "waveforms": {
            "wf_drag_gaussian_0": {
                "waveformId": "wf_drag_gaussian_0",
                "name": "drag_gaussian",
                "arguments": [
                    {"name": "length", "value": 6.000000000000001e-8, "type": "float"},
                    {"name": "sigma", "value": 6.369913502160144e-9, "type": "float"},
                    {"name": "amplitude", "value": -0.4549282253548838, "type": "float"},
                    {"name": "beta", "value": 7.494904522022295e-10, "type": "float"},
                ],
            },
        },
    }

    response_data_stream = io.BytesIO(json.dumps(response_data_content).encode("utf-8"))
    mock_url_request.return_value.__enter__.return_value = response_data_stream
    mock_session = Mock()
    mock_session.get_device.return_value = get_pulse_model(
        MOCK_PULSE_MODEL_QPU_PULSE_CAPABILITIES_JSON_1
    )
    device = AwsDevice(RIGETTI_ARN, mock_session)

    expected_waveforms = {
        "wf_drag_gaussian_0": DragGaussianWaveform(
            length=6.000000000000001e-8,
            sigma=6.369913502160144e-9,
            amplitude=-0.4549282253548838,
            beta=7.494904522022295e-10,
            id="wf_drag_gaussian_0",
        )
    }
    expected_ngc = GateCalibrations(
        pulse_sequences={
            (Gate.CPhaseShift(-1.5707963267948966), QubitSet([0, 1])): PulseSequence().play(
                device.frames["q0_q1_cphase_frame"], expected_waveforms["wf_drag_gaussian_0"]
            )
        }
    )
    assert device.gate_calibrations == expected_ngc
    # Called twice to check that the property stays the same after being initially fetched
    assert device.gate_calibrations == expected_ngc


@pytest.mark.xfail(raises=URLError)
@patch("urllib.request.urlopen")
def test_refresh_data_url_error(mock_url_request):
    mock_url_request.side_effect = URLError("mock reason")
    mock_session = Mock()
    mock_session.get_device.return_value = get_pulse_model(
        MOCK_PULSE_MODEL_QPU_PULSE_CAPABILITIES_JSON_1
    )
    device = AwsDevice(RIGETTI_ARN, mock_session)

    device.gate_calibrations


def test_equality(arn):
    mock_session = Mock()
    mock_session.get_device.return_value = MOCK_GATE_MODEL_QPU_1
    mock_session.region = RIGETTI_REGION
    device_1 = AwsDevice(arn, mock_session)
    device_2 = AwsDevice(arn, mock_session)
    other_device = AwsDevice("arn:aws:braket:::device/quantum-simulator/amazon/bar", mock_session)
    non_device = "HI"

    assert device_1 == device_2
    assert device_1 is not device_2
    assert device_1 != other_device
    assert device_1 != non_device


def test_repr(arn):
    mock_session = Mock()
    mock_session.get_device.return_value = MOCK_GATE_MODEL_QPU_1
    mock_session.region = RIGETTI_REGION
    device = AwsDevice(arn, mock_session)
    expected = "Device('name': {}, 'arn': {})".format(device.name, device.arn)
    assert repr(device) == expected


def test_device_simulator_not_found():
    mock_session = Mock()
    mock_session.region = "test-region-1"
    mock_session.get_device.side_effect = ClientError(
        {
            "Error": {
                "Code": "ResourceNotFoundException",
                "Message": (
                    "Braket device 'arn:aws:braket:::device/quantum-simulator/amazon/tn1' "
                    "not found in us-west-1. You can find a list of all supported device "
                    "ARNs and the regions in which they are available in the documentation: "
                    "https://docs.aws.amazon.com/braket/latest/developerguide/braket-devices.html"
                ),
            }
        },
        "getDevice",
    )
    simulator_not_found = (
        "Simulator 'arn:aws:braket:::device/simulator/a/b' not found in 'test-region-1'"
    )
    with pytest.raises(ValueError, match=simulator_not_found):
        AwsDevice("arn:aws:braket:::device/simulator/a/b", mock_session)


@patch("braket.aws.aws_device.AwsSession.copy_session")
def test_device_qpu_not_found(mock_copy_session):
    mock_session = Mock()
    mock_session.get_device.side_effect = ClientError(
        {
            "Error": {
                "Code": "ResourceNotFoundException",
                "Message": (
                    "Braket device 'arn:aws:braket:::device/quantum-simulator/amazon/tn1' "
                    "not found in us-west-1. You can find a list of all supported device "
                    "ARNs and the regions in which they are available in the documentation: "
                    "https://docs.aws.amazon.com/braket/latest/developerguide/braket-devices.html"
                ),
            }
        },
        "getDevice",
    )
    mock_copy_session.return_value = mock_session
    qpu_not_found = "QPU 'arn:aws:braket:::device/qpu/a/b' not found"
    with pytest.raises(ValueError, match=qpu_not_found):
        AwsDevice("arn:aws:braket:::device/qpu/a/b", mock_session)


@patch("braket.aws.aws_device.AwsSession.copy_session")
def test_device_qpu_exception(mock_copy_session):
    mock_session = Mock()
    mock_session.get_device.side_effect = (
        ClientError(
            {
                "Error": {
                    "Code": "ResourceNotFoundException",
                    "Message": (
                        "Braket device 'arn:aws:braket:::device/quantum-simulator/amazon/tn1' "
                        "not found in us-west-1. You can find a list of all supported device "
                        "ARNs and the regions in which they are available in the documentation: "
                        "https://docs.aws.amazon.com/braket/latest/developerguide/braket-"
                        "devices.html"
                    ),
                }
            },
            "getDevice",
        ),
        ClientError(
            {
                "Error": {
                    "Code": "OtherException",
                    "Message": "Some other message",
                }
            },
            "getDevice",
        ),
    )
    mock_copy_session.return_value = mock_session
    qpu_exception = (
        "An error occurred \\(OtherException\\) when calling the "
        "getDevice operation: Some other message"
    )
    with pytest.raises(ClientError, match=qpu_exception):
        AwsDevice("arn:aws:braket:::device/qpu/a/b", mock_session)


@patch("braket.aws.aws_device.AwsSession.copy_session")
def test_device_non_qpu_region_error(mock_copy_session):
    mock_session = Mock()
    mock_session.get_device.side_effect = ClientError(
        {
            "Error": {
                "Code": "ExpiredTokenError",
                "Message": ("Some other error that isn't ResourceNotFoundException"),
            }
        },
        "getDevice",
    )
    mock_copy_session.return_value = mock_session
    expired_token = (
        "An error occurred \\(ExpiredTokenError\\) when calling the getDevice operation: "
        "Some other error that isn't ResourceNotFoundException"
    )
    with pytest.raises(ClientError, match=expired_token):
        AwsDevice("arn:aws:braket:::device/qpu/a/b", mock_session)


@patch("braket.aws.aws_quantum_task.AwsQuantumTask.create")
def test_run_no_extra(aws_quantum_task_mock, device, circuit):
    _run_and_assert(
        aws_quantum_task_mock,
        device,
        circuit,
    )


@patch("braket.aws.aws_quantum_task.AwsQuantumTask.create")
def test_run_with_reservation_arn(aws_quantum_task_mock, device, circuit):
    _run_and_assert(
        aws_quantum_task_mock,
        device,
        circuit,
        reservation_arn="arn:aws:braket:us-west-2:123456789123:reservation/a1b123cd-45e6-789f-gh01-i234567jk8l9",
    )


@patch("braket.aws.aws_quantum_task.AwsQuantumTask")
def test_run_param_circuit_with_no_inputs(
    aws_quantum_task_mock, single_circuit_input, device, s3_destination_folder
):
    cannot_execute_with_unbound = "Cannot execute circuit with unbound parameters: {'theta'}"

    with pytest.raises(ValueError, match=cannot_execute_with_unbound):
        _run_and_assert(
            aws_quantum_task_mock,
            device,
            single_circuit_input,
            s3_destination_folder,
            10,
            86400,
            0.25,
            {},
        )


@patch("braket.aws.aws_quantum_task.AwsQuantumTask.create")
def test_run_param_circuit_with_inputs(
    aws_quantum_task_mock, single_circuit_input, device, s3_destination_folder
):
    inputs = {"theta": 0.2}

    _run_and_assert(
        aws_quantum_task_mock,
        device,
        single_circuit_input,
        s3_destination_folder,
        10,
        86400,
        0.25,
        inputs,
    )


@patch("braket.aws.aws_session.boto3.Session")
@patch("braket.aws.aws_session.AwsSession")
@patch("braket.aws.aws_quantum_task.AwsQuantumTask.create")
def test_run_param_circuit_with_reservation_arn_batch_task(
    aws_quantum_task_mock,
    aws_session_mock,
    boto_session_mock,
    single_circuit_input,
    device,
    s3_destination_folder,
):
    inputs = {"theta": 0.2}
    circ_1 = Circuit().rx(angle=0.2, target=0)
    circuits = [circ_1, single_circuit_input]

    _run_batch_and_assert(
        aws_quantum_task_mock,
        aws_session_mock,
        device,
        circuits,
        s3_destination_folder,
        10,
        20,
        50,
        43200,
        0.25,
        inputs,
        {},
        reservation_arn="arn:aws:braket:us-west-2:123456789123:reservation/a1b123cd-45e6-789f-gh01-i234567jk8l9",
    )


@patch("braket.aws.aws_session.boto3.Session")
@patch("braket.aws.aws_session.AwsSession")
@patch("braket.aws.aws_quantum_task.AwsQuantumTask.create")
def test_run_param_circuit_with_inputs_batch_task(
    aws_quantum_task_mock,
    aws_session_mock,
    boto_session_mock,
    single_circuit_input,
    device,
    s3_destination_folder,
):
    inputs = {"theta": 0.2}
    circ_1 = Circuit().rx(angle=0.2, target=0)
    circuits = [circ_1, single_circuit_input]

    _run_batch_and_assert(
        aws_quantum_task_mock,
        aws_session_mock,
        device,
        circuits,
        s3_destination_folder,
        10,
        20,
        50,
        43200,
        0.25,
        inputs,
        {},
    )


@patch("braket.aws.aws_quantum_task.AwsQuantumTask")
def test_run_param_circuit_with_invalid_input(
    aws_quantum_task_mock, single_circuit_input, device, s3_destination_folder
):
    inputs = {"beta": 0.2}
    cannot_execute_with_unbound = "Cannot execute circuit with unbound parameters: {'theta'}"
    with pytest.raises(ValueError, match=cannot_execute_with_unbound):
        _run_and_assert(
            aws_quantum_task_mock,
            device,
            single_circuit_input,
            s3_destination_folder,
            10,
            86400,
            0.25,
            inputs,
        )


@patch("braket.aws.aws_session.boto3.Session")
@patch("braket.aws.aws_session.AwsSession")
@patch("braket.aws.aws_quantum_task.AwsQuantumTask.create")
def test_run_batch_param_circuit_with_no_inputs(
    aws_quantum_task_mock,
    aws_session_mock,
    boto_session_mock,
    single_circuit_input,
    device,
    s3_destination_folder,
):
    circ_1 = Circuit().ry(angle=3, target=0)
    circuits = [circ_1, single_circuit_input]

    cannot_execute_with_unbound = "Cannot execute circuit with unbound parameters: {'theta'}"

    with pytest.raises(ValueError, match=cannot_execute_with_unbound):
        _run_batch_and_assert(
            aws_quantum_task_mock,
            aws_session_mock,
            device,
            circuits,
            s3_destination_folder,
            1000,
            20,
            50,
            43200,
            0.25,
            {},
        )


@patch("braket.aws.aws_session.boto3.Session")
@patch("braket.aws.aws_session.AwsSession")
@patch("braket.aws.aws_quantum_task_batch.AwsQuantumTask.create")
def test_run_multi_param_batch_circuit_with_input(
    aws_quantum_task_mock,
    aws_session_mock,
    boto_session_mock,
    multiple_circuit_inputs,
    device,
    s3_destination_folder,
):
    inputs = {"beta": 0.2}
    circ_1 = Circuit().ry(angle=3, target=0)
    circuits = [circ_1, multiple_circuit_inputs]

    cannot_execute_with_unbound = "Cannot execute circuit with unbound parameters: {'theta'}"
    with pytest.raises(ValueError, match=cannot_execute_with_unbound):
        _run_batch_and_assert(
            aws_quantum_task_mock,
            aws_session_mock,
            device,
            circuits,
            s3_destination_folder,
            1000,
            20,
            50,
            43200,
            0.25,
            inputs,
        )


@patch("braket.aws.aws_session.boto3.Session")
@patch("braket.aws.aws_session.AwsSession")
@patch("braket.aws.aws_quantum_task.AwsQuantumTask.create")
def test_run_param_batch_circuit_with_invalid_input(
    aws_quantum_task_mock,
    aws_session_mock,
    boto_session_mock,
    single_circuit_input,
    aws_session,
    device,
    s3_destination_folder,
):
    inputs = {"beta": 0.2}
    circ_1 = Circuit().ry(angle=3, target=0)
    circuits = [circ_1, single_circuit_input]
    cannot_execute_with_unbound = "Cannot execute circuit with unbound parameters: {'theta'}"
    with pytest.raises(ValueError, match=cannot_execute_with_unbound):
        _run_batch_and_assert(
            aws_quantum_task_mock,
            aws_session_mock,
            device,
            circuits,
            s3_destination_folder,
            1000,
            20,
            50,
            43200,
            0.25,
            inputs,
        )


@patch("braket.aws.aws_session.boto3.Session")
@patch("braket.aws.aws_session.AwsSession")
@patch("braket.aws.aws_quantum_task.AwsQuantumTask.create")
def test_batch_circuit_with_task_and_input_mismatch(
    aws_quantum_task_mock,
    aws_session_mock,
    boto_session_mock,
    single_circuit_input,
    openqasm_program,
    device,
    s3_destination_folder,
):
    inputs = [{"beta": 0.2}, {"gamma": 0.1}, {"theta": 0.2}]
    circ_1 = Circuit().ry(angle=3, target=0)
    task_specifications = [[circ_1, single_circuit_input], openqasm_program]
    wrong_number_of_inputs = (
        "Multiple inputs, task specifications and gate definitions must be equal in length."
    )

    with pytest.raises(ValueError, match=wrong_number_of_inputs):
        _run_batch_and_assert(
            aws_quantum_task_mock,
            aws_session_mock,
            device,
            task_specifications,
            s3_destination_folder,
            1000,
            20,
            50,
            43200,
            0.25,
            inputs,
            {},
        )


@patch("braket.aws.aws_session.boto3.Session")
@patch("braket.aws.aws_session.AwsSession")
@patch("braket.aws.aws_quantum_task.AwsQuantumTask.create")
def test_multiple_task_multiple_batch_inputs_invalid_config(
    aws_quantum_task_mock, aws_session_mock, boto_session_mock, device, s3_destination_folder
):
    theta = FreeParameter("theta")
    multiple_task = [Circuit().rx(angle=theta, target=1)] * 2
    multiple_inputs = [{"theta": 0.2}, {"beta": 0.3}]
    cannot_execute_with_unbound = "Cannot execute circuit with unbound parameters: {'theta'}"
    with pytest.raises(ValueError, match=cannot_execute_with_unbound):
        _run_batch_and_assert(
            aws_quantum_task_mock,
            aws_session_mock,
            device,
            multiple_task,
            s3_destination_folder,
            1000,
            20,
            50,
            43200,
            0.25,
            multiple_inputs,
        )


@patch("braket.aws.aws_session.boto3.Session")
@patch("braket.aws.aws_session.AwsSession")
@patch("braket.aws.aws_quantum_task.AwsQuantumTask.create")
def test_single_task_single_input_batch_missing_input(
    aws_quantum_task_mock, aws_session_mock, boto_session_mock, device, s3_destination_folder
):
    theta = FreeParameter("theta")
    task = Circuit().rx(angle=theta, target=0)
    inputs = {"beta": 0.2}
    cannot_execute_with_unbound = "Cannot execute circuit with unbound parameters: {'theta'}"
    with pytest.raises(ValueError, match=cannot_execute_with_unbound):
        _run_batch_and_assert(
            aws_quantum_task_mock,
            aws_session_mock,
            device,
            task,
            s3_destination_folder,
            1000,
            20,
            50,
            43200,
            0.25,
            inputs,
        )


@patch("braket.aws.aws_quantum_task.AwsQuantumTask.create")
def test_run_with_positional_args(aws_quantum_task_mock, device, circuit, s3_destination_folder):
    _run_and_assert(
        aws_quantum_task_mock,
        device,
        circuit,
        s3_destination_folder,
        100,
        86400,
        0.25,
        ["foo"],
    )


@patch("braket.aws.aws_quantum_task.AwsQuantumTask.create")
def test_run_with_kwargs(aws_quantum_task_mock, device, circuit, s3_destination_folder):
    _run_and_assert(
        aws_quantum_task_mock,
        device,
        circuit,
        s3_destination_folder,
        extra_kwargs={"bar": 1, "baz": 2},
    )


@patch("braket.aws.aws_quantum_task.AwsQuantumTask.create")
def test_run_with_shots(aws_quantum_task_mock, device, circuit, s3_destination_folder):
    _run_and_assert(aws_quantum_task_mock, device, circuit, s3_destination_folder, 100)


@patch("braket.aws.aws_quantum_task.AwsQuantumTask.create")
def test_run_with_shots_kwargs(aws_quantum_task_mock, device, circuit, s3_destination_folder):
    _run_and_assert(
        aws_quantum_task_mock,
        device,
        circuit,
        s3_destination_folder,
        100,
        extra_kwargs={"bar": 1, "baz": 2},
    )


@patch("braket.aws.aws_quantum_task.AwsQuantumTask.create")
def test_default_bucket_not_called(aws_quantum_task_mock, device, circuit, s3_destination_folder):
    device = device(RIGETTI_ARN)
    run_and_assert(
        aws_quantum_task_mock,
        device,
        MOCK_DEFAULT_S3_DESTINATION_FOLDER,
        AwsDevice.DEFAULT_SHOTS_QPU,
        AwsQuantumTask.DEFAULT_RESULTS_POLL_TIMEOUT,
        AwsQuantumTask.DEFAULT_RESULTS_POLL_INTERVAL,
        circuit,
        s3_destination_folder,
        shots=None,
        poll_timeout_seconds=None,
        poll_interval_seconds=None,
        inputs=None,
        gate_definitions=None,
        reservation_arn=None,
        extra_args=None,
        extra_kwargs=None,
    )
    device._aws_session.default_bucket.assert_not_called()


@patch("braket.aws.aws_quantum_task.AwsQuantumTask.create")
def test_run_device_poll_interval_kwargs(
    aws_quantum_task_mock, aws_session, circuit, s3_destination_folder
):
    poll_interval_seconds = 200
    capabilities = MOCK_GATE_MODEL_QPU_CAPABILITIES_1
    capabilities.service.getTaskPollIntervalMillis = poll_interval_seconds
    properties = {
        "deviceName": "Aspen-10",
        "deviceType": "QPU",
        "providerName": "provider1",
        "deviceStatus": "OFFLINE",
        "deviceCapabilities": capabilities.json(),
    }
    aws_session.get_device.return_value = properties
    _run_and_assert(
        aws_quantum_task_mock,
        lambda arn: AwsDevice(arn, aws_session),
        circuit,
        s3_destination_folder,
        100,
        86400,
        poll_interval_seconds / 1000.0,
        extra_kwargs={"bar": 1, "baz": 2},
    )


@patch("braket.aws.aws_quantum_task.AwsQuantumTask.create")
def test_run_with_shots_poll_timeout_kwargs(
    aws_quantum_task_mock, device, circuit, s3_destination_folder
):
    _run_and_assert(
        aws_quantum_task_mock,
        device,
        circuit,
        s3_destination_folder,
        100,
        86400,
        extra_kwargs={"bar": 1, "baz": 2},
    )


@patch("braket.aws.aws_quantum_task.AwsQuantumTask.create")
def test_run_with_positional_args_and_kwargs(
    aws_quantum_task_mock, device, circuit, s3_destination_folder
):
    _run_and_assert(
        aws_quantum_task_mock,
        device,
        circuit,
        s3_destination_folder,
        100,
        86400,
        0.25,
        {},
        {},
        "arn:aws:braket:us-west-2:123456789123:reservation/a1b123cd-45e6-789f-gh01-i234567jk8l9",
        None,
        {"bar": 1, "baz": 2},
    )


@patch.dict(
    os.environ,
    {"AMZN_BRAKET_TASK_RESULTS_S3_URI": "s3://env_bucket/env/path"},
)
@patch("braket.aws.aws_quantum_task.AwsQuantumTask.create")
def test_run_env_variables(aws_quantum_task_mock, device, circuit, arn):
    device(arn).run(circuit)
    assert aws_quantum_task_mock.call_args_list[0][0][3] == ("env_bucket", "env/path")


@patch("braket.aws.aws_session.boto3.Session")
@patch("braket.aws.aws_session.AwsSession")
@patch("braket.aws.aws_quantum_task.AwsQuantumTask.create")
def test_run_batch_no_extra(
    aws_quantum_task_mock,
    aws_session_mock,
    boto_session_mock,
    device,
    circuit,
    s3_destination_folder,
):
    _run_batch_and_assert(
        aws_quantum_task_mock,
        aws_session_mock,
        device,
        [circuit for _ in range(10)],
        s3_destination_folder,
        1000,
        20,
        50,
        43200,
        0.25,
        {},
        {},
    )


@patch("braket.aws.aws_session.boto3.Session")
@patch("braket.aws.aws_session.AwsSession")
@patch("braket.aws.aws_quantum_task.AwsQuantumTask.create")
def test_run_batch_with_shots(
    aws_quantum_task_mock,
    aws_session_mock,
    boto_session_mock,
    device,
    circuit,
    s3_destination_folder,
):
    _run_batch_and_assert(
        aws_quantum_task_mock,
        aws_session_mock,
        device,
        [circuit for _ in range(10)],
        s3_destination_folder,
        1000,
        20,
        50,
        43200,
        0.25,
        {},
        {},
    )


@patch("braket.aws.aws_session.boto3.Session")
@patch("braket.aws.aws_session.AwsSession")
@patch("braket.aws.aws_quantum_task.AwsQuantumTask.create")
def test_run_batch_with_max_parallel_and_kwargs(
    aws_quantum_task_mock,
    aws_session_mock,
    boto_session_mock,
    device,
    circuit,
    s3_destination_folder,
):
    _run_batch_and_assert(
        aws_quantum_task_mock,
        aws_session_mock,
        device,
        [circuit for _ in range(10)],
        s3_destination_folder,
        1000,
        20,
        50,
        43200,
        0.25,
        inputs={"theta": 0.2},
        gate_definitions={},
        extra_kwargs={"bar": 1, "baz": 2},
    )


@patch("braket.aws.aws_session.boto3.Session")
@patch.dict(
    os.environ,
    {"AMZN_BRAKET_TASK_RESULTS_S3_URI": "s3://env_bucket/env/path"},
)
@patch("braket.aws.aws_quantum_task.AwsQuantumTask.create")
def test_run_batch_env_variables(aws_quantum_task_mock, boto_session_mock, device, circuit, arn):
    device(arn).run_batch([circuit])
    assert aws_quantum_task_mock.call_args_list[0][0][3] == ("env_bucket", "env/path")


def _run_and_assert(
    aws_quantum_task_mock,
    device_factory,
    circuit,
    s3_destination_folder=None,  # Treated as positional arg
    shots=None,  # Treated as positional arg
    poll_timeout_seconds=None,  # Treated as positional arg
    poll_interval_seconds=None,  # Treated as positional arg
    inputs=None,  # Treated as positional arg
    gate_definitions=None,  # Treated as positional arg
    reservation_arn=None,  # Treated as positional arg
    extra_args=None,
    extra_kwargs=None,
):
    run_and_assert(
        aws_quantum_task_mock,
        device_factory("arn:aws:braket:::device/quantum-simulator/amazon/sim"),
        MOCK_DEFAULT_S3_DESTINATION_FOLDER,
        AwsDevice.DEFAULT_SHOTS_SIMULATOR,
        AwsQuantumTask.DEFAULT_RESULTS_POLL_TIMEOUT,
        AwsQuantumTask.DEFAULT_RESULTS_POLL_INTERVAL,
        circuit,
        s3_destination_folder,
        shots,
        poll_timeout_seconds,
        poll_interval_seconds,
        inputs,
        gate_definitions,
        reservation_arn,
        extra_args,
        extra_kwargs,
    )


def _run_batch_and_assert(
    aws_quantum_task_mock,
    aws_session_mock,
    device_factory,
    circuits,
    s3_destination_folder=None,  # Treated as positional arg
    shots=None,  # Treated as positional arg
    max_parallel=None,  # Treated as positional arg
    max_connections=None,  # Treated as positional arg
    poll_timeout_seconds=None,  # Treated as a positional arg
    poll_interval_seconds=None,  # Treated as positional arg
    inputs=None,  # Treated as positional arg
    gate_definitions=None,  # Treated as positional arg
    reservation_arn=None,  # Treated as positional arg
    extra_args=None,
    extra_kwargs=None,
):
    run_batch_and_assert(
        aws_quantum_task_mock,
        aws_session_mock,
        device_factory("arn:aws:braket:::device/quantum-simulator/amazon/sim"),
        MOCK_DEFAULT_S3_DESTINATION_FOLDER,
        AwsDevice.DEFAULT_SHOTS_SIMULATOR,
        AwsQuantumTask.DEFAULT_RESULTS_POLL_TIMEOUT,
        AwsQuantumTask.DEFAULT_RESULTS_POLL_INTERVAL,
        circuits,
        s3_destination_folder,
        shots,
        max_parallel,
        max_connections,
        poll_timeout_seconds,
        poll_interval_seconds,
        inputs,
        gate_definitions,
        reservation_arn,
        extra_args,
        extra_kwargs,
    )


def _assert_device_fields(device, expected_properties, expected_device_data):
    assert device.name == expected_device_data.get("deviceName")
    assert device.properties == expected_properties
    assert device.status == expected_device_data.get("deviceStatus")
    assert device.provider_name == expected_device_data.get("providerName")
    assert device.type == AwsDeviceType(expected_device_data.get("deviceType"))
    if device.topology_graph:
        assert device.topology_graph.edges == device._construct_topology_graph().edges
    assert device.frames == {}
    assert device.ports == {}


@patch("braket.aws.aws_device.AwsSession.copy_session")
def test_get_devices(mock_copy_session, aws_session):
    aws_session.search_devices.side_effect = [
        # us-west-1
        [
            {
                "deviceArn": SV1_ARN,
                "deviceName": "SV1",
                "deviceType": "SIMULATOR",
                "deviceStatus": "ONLINE",
                "providerName": "Amazon Braket",
            }
        ],
        ValueError("should not be reachable"),
    ]
    aws_session.get_device.side_effect = [
        MOCK_GATE_MODEL_SIMULATOR,
        ValueError("should not be reachable"),
    ]
    session_for_region = Mock()
    session_for_region.search_devices.side_effect = [
        # us-east-1
        [
            {
                "deviceArn": IONQ_ARN,
                "deviceName": "IonQ Device",
                "deviceType": "QPU",
                "deviceStatus": "ONLINE",
                "providerName": "IonQ",
            },
        ],
        # us-west-2
        [
            {
                "deviceArn": DWAVE_ARN,
                "deviceName": "Advantage_system1.1",
                "deviceType": "QPU",
                "deviceStatus": "ONLINE",
                "providerName": "D-Wave",
            },
            # Should not be reached because already instantiated in us-west-1
            {
                "deviceArn": SV1_ARN,
                "deviceName": "SV1",
                "deviceType": "SIMULATOR",
                "deviceStatus": "ONLINE",
                "providerName": "Amazon Braket",
            },
        ],
        # eu-west-2
        [
            {
                "deviceArn": OQC_ARN,
                "deviceName": "Lucy",
                "deviceType": "QPU",
                "deviceStatus": "ONLINE",
                "providerName": "OQC",
            }
        ],
        # Only two regions to search outside of current
        ValueError("should not be reachable"),
    ]
    session_for_region.get_device.side_effect = [
        MOCK_DWAVE_QPU,
        MOCK_GATE_MODEL_QPU_2,
        MOCK_GATE_MODEL_QPU_3,
        ValueError("should not be reachable"),
    ]
    mock_copy_session.return_value = session_for_region
    # Search order: us-east-1, us-west-1, us-west-2, eu-west-2
    results = AwsDevice.get_devices(
        arns=[SV1_ARN, DWAVE_ARN, IONQ_ARN, OQC_ARN],
        provider_names=["Amazon Braket", "D-Wave", "IonQ", "OQC"],
        statuses=["ONLINE"],
        aws_session=aws_session,
    )
    assert [result.name for result in results] == ["Advantage_system1.1", "Blah", "Lucy", "SV1"]


@patch("braket.aws.aws_device.AwsSession.copy_session")
def test_get_devices_simulators_only(mock_copy_session, aws_session):
    aws_session.search_devices.side_effect = [
        [
            {
                "deviceArn": SV1_ARN,
                "deviceName": "SV1",
                "deviceType": "SIMULATOR",
                "deviceStatus": "ONLINE",
                "providerName": "Amazon Braket",
            }
        ],
        ValueError("should not be reachable"),
    ]
    aws_session.get_device.side_effect = [
        MOCK_GATE_MODEL_SIMULATOR,
        ValueError("should not be reachable"),
    ]
    session_for_region = Mock()
    session_for_region.search_devices.side_effect = ValueError("should not be reachable")
    session_for_region.get_device.side_effect = ValueError("should not be reachable")
    mock_copy_session.return_value = session_for_region
    results = AwsDevice.get_devices(
        arns=[SV1_ARN, TN1_ARN],
        types=["SIMULATOR"],
        provider_names=["Amazon Braket"],
        statuses=["ONLINE"],
        aws_session=aws_session,
    )
    # Only one region should be searched
    assert [result.name for result in results] == ["SV1"]


@pytest.mark.filterwarnings("ignore:Test Code:")
@patch("braket.aws.aws_device.AwsSession.copy_session")
def test_get_devices_with_error_in_region(mock_copy_session, aws_session):
    aws_session.search_devices.side_effect = [
        # us-west-1
        [
            {
                "deviceArn": SV1_ARN,
                "deviceName": "SV1",
                "deviceType": "SIMULATOR",
                "deviceStatus": "ONLINE",
                "providerName": "Amazon Braket",
            }
        ],
        ValueError("should not be reachable"),
    ]
    aws_session.get_device.side_effect = [
        MOCK_GATE_MODEL_SIMULATOR,
        ValueError("should not be reachable"),
    ]
    session_for_region = Mock()
    session_for_region.search_devices.side_effect = [
        # us-east-1
        [
            {
                "deviceArn": IONQ_ARN,
                "deviceName": "IonQ Device",
                "deviceType": "QPU",
                "deviceStatus": "ONLINE",
                "providerName": "IonQ",
            },
        ],
        # us-west-2
        ClientError(
            {
                "Error": {
                    "Code": "Test Code",
                    "Message": "Test Message",
                }
            },
            "Test Operation",
        ),
        # eu-west-2
        [
            {
                "deviceArn": OQC_ARN,
                "deviceName": "Lucy",
                "deviceType": "QPU",
                "deviceStatus": "ONLINE",
                "providerName": "OQC",
            }
        ],
        # Only two regions to search outside of current
        ValueError("should not be reachable"),
    ]
    session_for_region.get_device.side_effect = [
        MOCK_GATE_MODEL_QPU_2,
        MOCK_GATE_MODEL_QPU_3,
        ValueError("should not be reachable"),
    ]
    mock_copy_session.return_value = session_for_region
    # Search order: us-east-1, us-west-1, us-west-2, eu-west-2
    results = AwsDevice.get_devices(
        statuses=["ONLINE"],
        aws_session=aws_session,
    )
    assert [result.name for result in results] == ["Blah", "Lucy", "SV1"]


@pytest.mark.xfail(raises=ValueError)
def test_get_devices_invalid_order_by():
    AwsDevice.get_devices(order_by="foo")


@patch("braket.aws.aws_device.datetime")
def test_get_device_availability(mock_utc_now):
    class Expando(object):
        pass

    class MockDevice(AwsDevice):
        def __init__(self, status, *execution_window_args):
            self._status = status
            self._properties = Expando()
            self._properties.service = Expando()
            execution_windows = []
            for execution_day, window_start_hour, window_end_hour in execution_window_args:
                execution_windows.append(
                    DeviceExecutionWindow.parse_raw(
                        json.dumps(
                            {
                                "executionDay": execution_day,
                                "windowStartHour": window_start_hour,
                                "windowEndHour": window_end_hour,
                            }
                        )
                    )
                )
            self._properties.service.executionWindows = execution_windows

    test_sets = (
        {
            "test_devices": (
                ("always_on_device", MockDevice("ONLINE", ("Everyday", "00:00", "23:59:59"))),
                ("offline_device", MockDevice("OFFLINE", ("Everyday", "00:00", "23:59:59"))),
                ("retired_device", MockDevice("RETIRED", ("Everyday", "00:00", "23:59:59"))),
                ("missing_schedule_device", MockDevice("ONLINE")),
            ),
            "test_items": (
                (datetime(2021, 12, 6, 10, 0, 0), (1, 0, 0, 0)),
                (datetime(2021, 12, 7, 10, 0, 0), (1, 0, 0, 0)),
                (datetime(2021, 12, 8, 10, 0, 0), (1, 0, 0, 0)),
                (datetime(2021, 12, 9, 10, 0, 0), (1, 0, 0, 0)),
                (datetime(2021, 12, 10, 10, 0, 0), (1, 0, 0, 0)),
                (datetime(2021, 12, 11, 10, 0, 0), (1, 0, 0, 0)),
                (datetime(2021, 12, 12, 10, 0, 0), (1, 0, 0, 0)),
            ),
        },
        {
            "test_devices": (
                ("midday_everyday_device", MockDevice("ONLINE", ("Everyday", "07:00", "17:00"))),
                ("midday_weekday_device", MockDevice("ONLINE", ("Weekdays", "07:00", "17:00"))),
                ("midday_weekend_device", MockDevice("ONLINE", ("Weekend", "07:00", "17:00"))),
                ("evening_everyday_device", MockDevice("ONLINE", ("Everyday", "17:00", "07:00"))),
                ("evening_weekday_device", MockDevice("ONLINE", ("Weekdays", "17:00", "07:00"))),
                ("evening_weekend_device", MockDevice("ONLINE", ("Weekend", "17:00", "07:00"))),
            ),
            "test_items": (
                (datetime(2021, 12, 6, 5, 0, 0), (0, 0, 0, 1, 0, 1)),
                (datetime(2021, 12, 6, 10, 0, 0), (1, 1, 0, 0, 0, 0)),
                (datetime(2021, 12, 6, 20, 0, 0), (0, 0, 0, 1, 1, 0)),
                (datetime(2021, 12, 7, 5, 0, 0), (0, 0, 0, 1, 1, 0)),
                (datetime(2021, 12, 7, 10, 0, 0), (1, 1, 0, 0, 0, 0)),
                (datetime(2021, 12, 7, 20, 0, 0), (0, 0, 0, 1, 1, 0)),
                (datetime(2021, 12, 8, 5, 0, 0), (0, 0, 0, 1, 1, 0)),
                (datetime(2021, 12, 8, 10, 0, 0), (1, 1, 0, 0, 0, 0)),
                (datetime(2021, 12, 8, 20, 0, 0), (0, 0, 0, 1, 1, 0)),
                (datetime(2021, 12, 9, 5, 0, 0), (0, 0, 0, 1, 1, 0)),
                (datetime(2021, 12, 9, 10, 0, 0), (1, 1, 0, 0, 0, 0)),
                (datetime(2021, 12, 9, 20, 0, 0), (0, 0, 0, 1, 1, 0)),
                (datetime(2021, 12, 10, 5, 0, 0), (0, 0, 0, 1, 1, 0)),
                (datetime(2021, 12, 10, 10, 0, 0), (1, 1, 0, 0, 0, 0)),
                (datetime(2021, 12, 10, 20, 0, 0), (0, 0, 0, 1, 1, 0)),
                (datetime(2021, 12, 11, 5, 0, 0), (0, 0, 0, 1, 1, 0)),
                (datetime(2021, 12, 11, 10, 0, 0), (1, 0, 1, 0, 0, 0)),
                (datetime(2021, 12, 11, 20, 0, 0), (0, 0, 0, 1, 0, 1)),
                (datetime(2021, 12, 12, 5, 0, 0), (0, 0, 0, 1, 0, 1)),
                (datetime(2021, 12, 12, 10, 0, 0), (1, 0, 1, 0, 0, 0)),
                (datetime(2021, 12, 12, 20, 0, 0), (0, 0, 0, 1, 0, 1)),
            ),
        },
        {
            "test_devices": (
                ("monday_device", MockDevice("ONLINE", ("Monday", "07:00", "17:00"))),
                ("tuesday_device", MockDevice("ONLINE", ("Tuesday", "07:00", "17:00"))),
                ("wednesday_device", MockDevice("ONLINE", ("Wednesday", "07:00", "17:00"))),
                ("thursday_device", MockDevice("ONLINE", ("Thursday", "07:00", "17:00"))),
                ("friday_device", MockDevice("ONLINE", ("Friday", "07:00", "17:00"))),
                ("saturday_device", MockDevice("ONLINE", ("Saturday", "07:00", "17:00"))),
                ("sunday_device", MockDevice("ONLINE", ("Sunday", "07:00", "17:00"))),
                (
                    "monday_friday_device",
                    MockDevice(
                        "ONLINE", ("Monday", "07:00", "17:00"), ("Friday", "07:00", "17:00")
                    ),
                ),
            ),
            "test_items": (
                (datetime(2021, 12, 6, 5, 0, 0), (0, 0, 0, 0, 0, 0, 0, 0)),
                (datetime(2021, 12, 6, 10, 0, 0), (1, 0, 0, 0, 0, 0, 0, 1)),
                (datetime(2021, 12, 6, 20, 0, 0), (0, 0, 0, 0, 0, 0, 0, 0)),
                (datetime(2021, 12, 7, 5, 0, 0), (0, 0, 0, 0, 0, 0, 0, 0)),
                (datetime(2021, 12, 7, 10, 0, 0), (0, 1, 0, 0, 0, 0, 0, 0)),
                (datetime(2021, 12, 7, 20, 0, 0), (0, 0, 0, 0, 0, 0, 0, 0)),
                (datetime(2021, 12, 8, 5, 0, 0), (0, 0, 0, 0, 0, 0, 0, 0)),
                (datetime(2021, 12, 8, 10, 0, 0), (0, 0, 1, 0, 0, 0, 0, 0)),
                (datetime(2021, 12, 8, 20, 0, 0), (0, 0, 0, 0, 0, 0, 0, 0)),
                (datetime(2021, 12, 9, 5, 0, 0), (0, 0, 0, 0, 0, 0, 0, 0)),
                (datetime(2021, 12, 9, 10, 0, 0), (0, 0, 0, 1, 0, 0, 0, 0)),
                (datetime(2021, 12, 9, 20, 0, 0), (0, 0, 0, 0, 0, 0, 0, 0)),
                (datetime(2021, 12, 10, 5, 0, 0), (0, 0, 0, 0, 0, 0, 0, 0)),
                (datetime(2021, 12, 10, 10, 0, 0), (0, 0, 0, 0, 1, 0, 0, 1)),
                (datetime(2021, 12, 10, 20, 0, 0), (0, 0, 0, 0, 0, 0, 0, 0)),
                (datetime(2021, 12, 11, 5, 0, 0), (0, 0, 0, 0, 0, 0, 0, 0)),
                (datetime(2021, 12, 11, 10, 0, 0), (0, 0, 0, 0, 0, 1, 0, 0)),
                (datetime(2021, 12, 11, 20, 0, 0), (0, 0, 0, 0, 0, 0, 0, 0)),
                (datetime(2021, 12, 12, 5, 0, 0), (0, 0, 0, 0, 0, 0, 0, 0)),
                (datetime(2021, 12, 12, 10, 0, 0), (0, 0, 0, 0, 0, 0, 1, 0)),
                (datetime(2021, 12, 12, 20, 0, 0), (0, 0, 0, 0, 0, 0, 0, 0)),
            ),
        },
    )

    for test_set in test_sets:
        for test_item in test_set["test_items"]:
            test_date = test_item[0]
            mock_utc_now.utcnow.return_value = test_date

            # flake8: noqa: C501
            for i in range(len(test_item[1])):
                device_name = test_set["test_devices"][i][0]
                device = test_set["test_devices"][i][1]
                type(device).properties = PropertyMock(return_value=Expando())
                type(device).properties.service = PropertyMock(return_value=Expando())
                device.properties.service.executionWindows = (
                    device._properties.service.executionWindows
                )
                expected = bool(test_item[1][i])
                actual = device.is_available
                assert (
                    expected == actual
                ), f"device_name: {device_name}, test_date: {test_date}, expected: {expected}, actual: {actual}"


@pytest.mark.parametrize(
    "get_device_data, expected_graph",
    [
        (MOCK_GATE_MODEL_QPU_1, nx.DiGraph([(1, 2), (1, 3)])),
        (MOCK_GATE_MODEL_QPU_2, nx.complete_graph(30, nx.DiGraph())),
        (MOCK_DWAVE_QPU, nx.DiGraph([(1, 2), (2, 3)])),
    ],
)
def test_device_topology_graph_data(get_device_data, expected_graph, arn):
    mock_session = Mock()
    mock_session.get_device.return_value = get_device_data
    mock_session.region = RIGETTI_REGION
    device = AwsDevice(arn, mock_session)
    assert nx.is_isomorphic(device.topology_graph, expected_graph)
    new_val = "new_val"
    device._topology_graph = new_val
    assert device.topology_graph == new_val


def test_device_no_href():
    mock_session = Mock()
    mock_session.get_device.return_value = MOCK_GATE_MODEL_QPU_1
    device = AwsDevice(DWAVE_ARN, mock_session)


def test_parse_calibration_data():
    mock_session = Mock()
    mock_session.get_device.return_value = get_pulse_model(
        MOCK_PULSE_MODEL_QPU_PULSE_CAPABILITIES_JSON_1
    )
    device = AwsDevice(DWAVE_ARN, mock_session)
    calibration_data = device._parse_calibration_json(MOCK_gate_calibrations_JSON)
    device_ngc = GateCalibrations(calibration_data)

    expected_waveforms = {
        "wf_drag_gaussian_0": DragGaussianWaveform(
            length=6.000000000000001e-8,
            sigma=6.369913502160144e-9,
            amplitude=-0.4549282253548838,
            beta=7.494904522022295e-10,
            id="wf_drag_gaussian_0",
        )
    }
    expected_pulse_sequences = {
        (Gate.CPhaseShift(-1.5707963267948966), QubitSet([0, 1])): PulseSequence()
        .barrier(QubitSet(0))
        .play(device.frames["q0_q1_cphase_frame"], expected_waveforms["wf_drag_gaussian_0"])
        .barrier([device.frames["q0_q1_cphase_frame"]])
        .barrier([])
        .delay(QubitSet([0, 1]), 3e-07)
        .delay([device.frames["q0_q1_cphase_frame"]], 3e-07)
        .shift_phase(device.frames["q0_q1_cphase_frame"], 3e-07)
        .shift_frequency(device.frames["q0_q1_cphase_frame"], FreeParameter("theta")),
        (Gate.Rx(FreeParameter("theta")), QubitSet(0)): PulseSequence().barrier([]),
        (Gate.CZ(), QubitSet([1, 0])): PulseSequence().barrier([]),
    }
    expected_ngc = GateCalibrations(pulse_sequences=expected_pulse_sequences)
    assert device_ngc == expected_ngc


@pytest.mark.parametrize(
    "bad_input",
    [
        (
            {
                "gates": {
                    "0": {
                        "rx": [
                            {
                                "name": "rx",
                                "qubits": ["0"],
                                "arguments": ["-1.5707963267948966"],
                                "calibrations": [
                                    {
                                        "name": "incorrect_instr",
                                        "arguments": [
                                            {"name": "qubit", "value": "0", "type": "string"}
                                        ],
                                    }
                                ],
                            }
                        ]
                    }
                },
                "waveforms": {},
            }
        ),
        (
            {
                "gates": {
                    "0": {
                        "rx": [
                            {
                                "name": "cphaseshift",
                                "qubits": ["0"],
                                "arguments": ["-1.5707963267948966"],
                                "calibrations": [
                                    {
                                        "name": "delay",
                                        "arguments": [
                                            {"name": "bad_value", "value": "1", "type": "float"},
                                            {"name": "qubit", "value": None, "type": "string"},
                                        ],
                                    }
                                ],
                            }
                        ]
                    }
                },
                "waveforms": {
                    "blankId_waveform": {"waveformId": "blankId_waveform", "name": "bad_waveform"},
                },
            }
        ),
    ],
)
@pytest.mark.xfail(raises=ValueError)
def test_parse_calibration_data_bad_instr(bad_input):
    mock_session = Mock()
    mock_session.get_device.return_value = get_pulse_model(
        MOCK_PULSE_MODEL_QPU_PULSE_CAPABILITIES_JSON_1
    )
    device = AwsDevice(DWAVE_ARN, mock_session)
    device._parse_calibration_json(bad_input)


def test_queue_depth(arn):
    mock_session = Mock()
    mock_session.get_device.return_value = MOCK_GATE_MODEL_QPU_1
    mock_session.region = RIGETTI_REGION
    device = AwsDevice(arn, mock_session)
    assert device.queue_depth() == QueueDepthInfo(
        quantum_tasks={QueueType.NORMAL: "19", QueueType.PRIORITY: "3"},
        jobs="0 (3 prioritized job(s) running)",
    )


@pytest.fixture
def noise_model():
    return (
        NoiseModel()
        .add_noise(Noise.BitFlip(0.05), GateCriteria(Gate.H))
        .add_noise(Noise.TwoQubitDepolarizing(0.10), GateCriteria(Gate.CNot))
    )


@patch.dict(
    os.environ,
    {"AMZN_BRAKET_TASK_RESULTS_S3_URI": "s3://env_bucket/env/path"},
)
@patch("braket.aws.aws_device.AwsSession")
@patch("braket.aws.aws_quantum_task.AwsQuantumTask.create")
def test_run_with_noise_model(aws_quantum_task_mock, aws_session_init, aws_session, noise_model):
    arn = DM1_ARN
    aws_session_init.return_value = aws_session
    aws_session.get_device.return_value = MOCK_GATE_MODEL_NOISE_SIMULATOR
    device = AwsDevice(arn, noise_model=noise_model)
    circuit = Circuit().h(0).cnot(0, 1)
    _ = device.run(circuit)

    expected_circuit = textwrap.dedent(
        """
        OPENQASM 3.0;
        bit[2] b;
        qubit[2] q;
        h q[0];
        #pragma braket noise bit_flip(0.05) q[0]
        cnot q[0], q[1];
        #pragma braket noise two_qubit_depolarizing(0.1) q[0], q[1]
        b[0] = measure q[0];
        b[1] = measure q[1];
        """
    ).strip()

    expected_circuit = Circuit().h(0).bit_flip(0, 0.05).cnot(0, 1).two_qubit_depolarizing(0, 1, 0.1)
    assert aws_quantum_task_mock.call_args_list[0][0][2] == expected_circuit


@patch.dict(
    os.environ,
    {"AMZN_BRAKET_TASK_RESULTS_S3_URI": "s3://env_bucket/env/path"},
)
@patch("braket.aws.aws_device.AwsSession")
@patch("braket.aws.aws_quantum_task.AwsQuantumTask.create")
def test_run_batch_with_noise_model(
    aws_quantum_task_mock, aws_session_init, aws_session, noise_model
):
    arn = DM1_ARN
    aws_session_init.return_value = aws_session
    aws_session.get_device.return_value = MOCK_GATE_MODEL_NOISE_SIMULATOR
    device = AwsDevice(arn, noise_model=noise_model)
    circuit = Circuit().h(0).cnot(0, 1)
    _ = device.run_batch([circuit] * 2)

    expected_circuit = textwrap.dedent(
        """
        OPENQASM 3.0;
        bit[2] b;
        qubit[2] q;
        h q[0];
        #pragma braket noise bit_flip(0.05) q[0]
        cnot q[0], q[1];
        #pragma braket noise two_qubit_depolarizing(0.1) q[0], q[1]
        b[0] = measure q[0];
        b[1] = measure q[1];
        """
    ).strip()

    expected_circuit = Circuit().h(0).bit_flip(0, 0.05).cnot(0, 1).two_qubit_depolarizing(0, 1, 0.1)
    assert aws_quantum_task_mock.call_args_list[0][0][2] == expected_circuit



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from collections import OrderedDict

import pytest

from braket.circuits import Gate, Instruction, Moments, MomentsKey, QubitSet


def cnot(q1, q2):
    return Instruction(Gate.CNot(), [q1, q2])


def h(q):
    return Instruction(Gate.H(), q)


@pytest.fixture
def moments():
    return Moments([h(0), h(0)])


def test_add():
    moments = Moments()
    moments.add([h(0)])
    moments.add([h(0)])

    expected = OrderedDict()
    expected[MomentsKey(0, QubitSet(0), "gate", 0)] = h(0)
    expected[MomentsKey(1, QubitSet(0), "gate", 0)] = h(0)
    assert OrderedDict(moments) == expected


def test_add_single_insturction():
    moments = Moments()
    moments.add(h(0))

    expected = OrderedDict()
    expected[MomentsKey(0, QubitSet(0), "gate", 0)] = h(0)
    assert OrderedDict(moments) == expected


def test_default_constructor():
    moments = Moments()
    assert OrderedDict(moments) == OrderedDict()
    assert moments.depth == 0
    assert moments.qubits == QubitSet()


def test_constructor_with_instructions():
    moments = Moments([h(0), h(1)])
    expected = Moments()
    expected.add([h(0)])
    expected.add([h(1)])
    assert moments == expected


def test_depth():
    moments = Moments([h(0), h(1)])
    assert moments.depth == 1

    moments.add([cnot(0, 2), h(3)])
    assert moments.depth == 2


def test_depth_setter(moments):
    with pytest.raises(AttributeError):
        moments.depth = 5


def test_overlaping_qubits():
    moments = Moments([h(0), h(0)])
    assert moments.depth == 2

    moments.add([cnot(0, 3), h(1)])
    assert moments.depth == 3

    moments.add([cnot(2, 4)])
    assert moments.depth == 3


def test_qubits():
    moments = Moments([h(0), h(10), h(5)])
    expected = QubitSet([0, 10, 5])
    assert moments.qubits == expected
    assert moments.qubit_count == len(expected)


def test_qubits_setter(moments):
    with pytest.raises(AttributeError):
        moments.qubits = QubitSet(1)


def test_qubit_count_setter(moments):
    with pytest.raises(AttributeError):
        moments.qubit_count = 1


def test_time_slices():
    moments = Moments([h(0), h(1), cnot(0, 1)])
    expected = {0: [h(0), h(1)], 1: [cnot(0, 1)]}
    assert moments.time_slices() == expected


def test_keys():
    moments = Moments([h(0), h(0), h(1)])
    expected = [
        MomentsKey(0, QubitSet(0), "gate", 0),
        MomentsKey(1, QubitSet(0), "gate", 0),
        MomentsKey(0, QubitSet(1), "gate", 0),
    ]
    assert list(moments.keys()) == expected


def test_items():
    moments = Moments([h(0), h(0), h(1)])
    expected = [
        (MomentsKey(0, QubitSet(0), "gate", 0), h(0)),
        (MomentsKey(1, QubitSet(0), "gate", 0), h(0)),
        (MomentsKey(0, QubitSet(1), "gate", 0), h(1)),
    ]
    assert list(moments.items()) == expected


def test_values():
    moments = Moments([h(0), h(0), h(1)])
    expected = [h(0), h(0), h(1)]
    assert list(moments.values()) == expected


def test_get():
    moments = Moments([h(0)])
    unknown_key = MomentsKey(100, QubitSet(100), "gate", 0)
    assert moments.get(MomentsKey(0, QubitSet(0), "gate", 0)) == h(0)
    assert moments.get(unknown_key) is None
    assert moments.get(unknown_key, h(0)) == h(0)


def test_getitem():
    moments = Moments([h(0)])
    assert moments[MomentsKey(0, QubitSet(0), "gate", 0)] == h(0)


def test_iter(moments):
    assert [key for key in moments] == list(moments.keys())


def test_len():
    moments = Moments([h(0), h(0)])
    assert len(moments) == 2


def test_contains():
    moments = Moments([h(0), h(0)])
    assert MomentsKey(0, QubitSet(0), "gate", 0) in moments
    assert MomentsKey(0, QubitSet(100), "gate", 0) not in moments


def test_equals():
    moments_1 = Moments([h(0)])
    moments_2 = Moments([h(0)])
    other_moments = Moments([h(1)])
    non_moments = "non moments"

    assert moments_1 == moments_2
    assert moments_1 is not moments_2
    assert moments_1 != other_moments
    assert moments_1 != non_moments


def test_repr(moments):
    assert repr(moments) == repr(OrderedDict(moments))


def test_str(moments):
    assert str(moments) == str(OrderedDict(moments))



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import numpy as np
import pytest

from braket.registers import Qubit


@pytest.fixture
def qubit():
    return Qubit(5)


def test_index_lt_zero():
    with pytest.raises(ValueError):
        Qubit(-1)


@pytest.mark.parametrize("qubit_arg", ("not a number", 0.5))
def test_index_non_int(qubit_arg):
    with pytest.raises(TypeError):
        Qubit(qubit_arg)


@pytest.mark.parametrize("qubit_index", (0, 5, np.int64(5)))
def test_index_gte_zero(qubit_index):
    Qubit(qubit_index)


def test_str(qubit):
    expected = "Qubit({})".format(int(qubit))
    assert str(qubit) == expected


def test_new_with_qubit():
    qubit = Qubit(0)
    qubit_new = Qubit.new(qubit)
    assert qubit_new == qubit
    assert qubit_new is qubit


def test_new_with_int():
    qubit = 0
    qubit_new = Qubit.new(qubit)
    assert qubit_new == qubit
    assert qubit_new is not qubit



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import pytest

from braket.parametric import FreeParameter, FreeParameterExpression
from braket.parametric.free_parameter_expression import subs_if_free_parameter


@pytest.fixture
def free_parameter_expression():
    return FreeParameterExpression(FreeParameter("theta") - 1)


def test_is_free_param_expr(free_parameter_expression):
    assert isinstance(free_parameter_expression, FreeParameterExpression)


@pytest.mark.xfail(raises=NotImplementedError)
def test_constructor_bad_input():
    FreeParameterExpression(["test"])


def test_equality():
    expr_1 = FreeParameterExpression(FreeParameter("theta") + 1)
    expr_2 = FreeParameterExpression(FreeParameter("theta") + 1)
    other_expr = FreeParameterExpression(FreeParameter("theta"))
    non_expr = "non circuit"

    assert expr_1 == expr_2
    assert expr_1 is not expr_2
    assert expr_1 != other_expr
    assert expr_1 != non_expr


def test_equality_str():
    expr_1 = FreeParameterExpression("-theta+2*theta")
    expr_2 = FreeParameterExpression(-FreeParameter("theta") + 2 * FreeParameter("theta"))
    param_values = {"theta": 1}
    assert expr_1 == expr_2
    assert expr_1.subs(param_values) == expr_2.subs(param_values)
    assert hasattr(expr_1.expression, "free_symbols") and hasattr(expr_2.expression, "free_symbols")


@pytest.mark.xfail(raises=ValueError)
def test_unsupported_bin_op_str():
    FreeParameterExpression("theta/1")


@pytest.mark.xfail(raises=ValueError)
def test_unsupported_un_op_str():
    FreeParameterExpression("~theta")


@pytest.mark.xfail(raises=ValueError)
def test_unsupported_node_str():
    FreeParameterExpression("theta , 1")


def test_commutativity():
    add_1 = 1 + FreeParameterExpression(FreeParameter("theta"))
    add_2 = FreeParameterExpression(FreeParameter("theta")) + 1
    mul_1 = FreeParameterExpression(FreeParameter("theta") * 1)
    mul_2 = FreeParameterExpression(1 * FreeParameter("theta"))

    assert add_1 == add_2
    assert mul_1 == mul_2


def test_add():
    add_expr = FreeParameter("theta") + FreeParameter("theta")
    expected = FreeParameterExpression(2 * FreeParameter("theta"))
    assert add_expr == expected


def test_sub():
    sub_expr = FreeParameter("theta") - FreeParameter("alpha")
    expected = FreeParameterExpression(FreeParameter("theta")) - FreeParameterExpression(
        FreeParameter("alpha")
    )
    assert sub_expr == expected


def test_r_sub():
    r_sub_expr = 1 - FreeParameter("theta")
    expected = FreeParameterExpression(1 - FreeParameter("theta"))
    assert r_sub_expr == expected


def test_mul():
    mul_expr = FreeParameter("theta") * FreeParameter("alpha") * 2 * FreeParameter("theta")
    expected = FreeParameterExpression(FreeParameter("theta") ** 2 * FreeParameter("alpha") * 2)
    assert mul_expr == expected


def test_truediv():
    truediv_expr = FreeParameter("theta") / FreeParameter("alpha")
    expected = FreeParameterExpression(FreeParameter("theta")) / FreeParameterExpression(
        FreeParameter("alpha")
    )
    assert truediv_expr == expected


def test_r_truediv():
    r_truediv_expr = 1 / FreeParameter("theta")
    expected = FreeParameterExpression(1 / FreeParameter("theta"))
    assert r_truediv_expr == expected


def test_pow():
    mul_expr = FreeParameter("theta") ** FreeParameter("alpha") * 2
    expected = FreeParameterExpression(FreeParameter("theta") ** FreeParameter("alpha") * 2)
    assert mul_expr == expected


def test_pow_constant():
    mul_expr = FreeParameter("theta") ** 2
    expected = FreeParameterExpression(FreeParameter("theta") ** 2)
    assert mul_expr == expected


def test_r_pow():
    mul_expr = 2 ** FreeParameter("theta")
    expected = FreeParameterExpression(2 ** FreeParameter("theta"))
    assert mul_expr == expected


def test_neg():
    expr = FreeParameter("theta") * FreeParameter("alpha") * 2
    expected_expr = -FreeParameter("theta") * -FreeParameter("alpha") * -2
    assert -expr == expected_expr and -(-expr) == expr


def test_sub_string():
    theta = FreeParameter("theta")
    expr = theta + 1
    assert expr.subs({"theta": 1}) == 2


def test_sub_free_parameter():
    theta = FreeParameter("theta")
    expr = theta + 1
    param_values = {theta: 1}
    assert expr.subs(param_values) == 2


def test_sub_return_expression():
    expr = FreeParameter("theta") + 1 + FreeParameter("alpha")
    subbed_expr = expr.subs({"alpha": 1})
    expected = FreeParameter("theta") + 2

    assert subbed_expr == expected


@pytest.mark.parametrize(
    "param, kwargs, expected_value, expected_type",
    [
        (FreeParameter("a") + 2 * FreeParameter("b"), {"a": 0.1, "b": 0.3}, 0.7, float),
        (FreeParameter("x"), {"y": 1}, FreeParameter("x"), FreeParameter),
        (FreeParameter("y"), {"y": -0.1}, -0.1, float),
        (2 * FreeParameter("i"), {"i": 1}, 2.0, float),
        (
            FreeParameter("a") + 2 * FreeParameter("x"),
            {"a": 0.4, "b": 0.4},
            0.4 + 2 * FreeParameter("x"),
            FreeParameterExpression,
        ),
    ],
)
def test_subs_if_free_parameter(param, kwargs, expected_value, expected_type):
    value = subs_if_free_parameter(param, **kwargs)
    assert value == expected_value
    assert isinstance(value, expected_type)



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from braket.aws import AwsDevice
from braket.circuits import Circuit
from braket.devices import Devices

bell = Circuit().h(0).cnot(0, 1)
device = AwsDevice(Devices.IonQ.Aria1)

# To run a task in a device reservation, change the device to the one you reserved
# and fill in your reservation ARN
task = device.run(bell, shots=100, reservation_arn="reservation ARN")
print(task.result().measurement_counts)



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import annotations

import json
from collections import Counter
from collections.abc import Callable
from dataclasses import dataclass
from typing import Any, Optional, TypeVar, Union

import numpy as np

from braket.circuits import Observable, ResultType, StandardObservable
from braket.circuits.observables import TensorProduct, observable_from_ir
from braket.ir.jaqcd import Expectation, Probability, Sample, Variance
from braket.task_result import (
    AdditionalMetadata,
    GateModelTaskResult,
    ResultTypeValue,
    TaskMetadata,
)

T = TypeVar("T")


@dataclass
class GateModelQuantumTaskResult:
    """Result of a gate model quantum task execution. This class is intended
    to be initialized by a QuantumTask class.

    Args:
        task_metadata (TaskMetadata): Quantum task metadata.
        additional_metadata (AdditionalMetadata): Additional metadata about the quantum task
        result_types (list[dict[str, Any]]): List of dictionaries where each dictionary
            has two keys: 'Type' (the result type in IR JSON form) and
            'Value' (the result value for this result type).
            This can be an empty list if no result types are specified in the IR.
            This is calculated from `measurements` and
            the IR of the circuit program when `shots>0`.
        values (list[Any]): The values for result types requested in the circuit.
            This can be an empty list if no result types are specified in the IR.
            This is calculated from `measurements` and
            the IR of the circuit program when `shots>0`.
        measurements (numpy.ndarray, optional): 2d array - row is shot and column is qubit.
            Default is None. Only available when shots > 0. The qubits in `measurements`
            are the ones in `GateModelQuantumTaskResult.measured_qubits`.
        measured_qubits (list[int], optional): The indices of the measured qubits. Default
            is None. Only available when shots > 0. Indicates which qubits are in
            `measurements`.
        measurement_counts (Counter, optional): A `Counter` of measurements. Key is the measurements
            in a big endian binary string. Value is the number of times that measurement occurred.
            Default is None. Only available when shots > 0. Note that the keys in `Counter` are
            unordered.
        measurement_probabilities (dict[str, float], optional):
            A dictionary of probabilistic results.
            Key is the measurements in a big endian binary string.
            Value is the probability the measurement occurred.
            Default is None. Only available when shots > 0.
        measurements_copied_from_device (bool, optional): flag whether `measurements`
            were copied from device. If false, `measurements` are calculated from device data.
            Default is None. Only available when shots > 0.
        measurement_counts_copied_from_device (bool, optional): flag whether `measurement_counts`
            were copied from device. If False, `measurement_counts` are calculated from device data.
            Default is None. Only available when shots > 0.
        measurement_probabilities_copied_from_device (bool, optional): flag whether
            `measurement_probabilities` were copied from device. If false,
            `measurement_probabilities` are calculated from device data. Default is None.
            Only available when shots > 0.
    """

    task_metadata: TaskMetadata
    additional_metadata: AdditionalMetadata
    result_types: list[ResultTypeValue] = None
    values: list[Any] = None
    measurements: np.ndarray = None
    measured_qubits: list[int] = None
    measurement_counts: Counter = None
    measurement_probabilities: dict[str, float] = None
    measurements_copied_from_device: bool = None
    measurement_counts_copied_from_device: bool = None
    measurement_probabilities_copied_from_device: bool = None

    _result_types_indices: dict[str, int] = None

    def __post_init__(self):
        if self.result_types is not None:
            self._result_types_indices = {
                GateModelQuantumTaskResult._result_type_hash(rt.type): i
                for i, rt in enumerate(self.result_types)
            }
        else:
            self._result_types_indices = {}

    def get_value_by_result_type(self, result_type: ResultType) -> Any:
        """Get value by result type. The result type must have already been
        requested in the circuit sent to the device for this quantum task result.

        Args:
            result_type (ResultType): result type requested

        Returns:
            Any: value of the result corresponding to the result type

        Raises:
            ValueError: If result type is not found in result.
                Result types must be added to the circuit before the circuit is run on a device.
        """
        rt_ir = result_type.to_ir()
        try:
            rt_hash = GateModelQuantumTaskResult._result_type_hash(rt_ir)
            result_type_index = self._result_types_indices[rt_hash]
            return self.values[result_type_index]
        except KeyError:
            raise ValueError(
                "Result type not found in result. "
                "Result types must be added to circuit before circuit is run on device."
            )

    def __eq__(self, other: GateModelQuantumTaskResult) -> bool:
        if isinstance(other, GateModelQuantumTaskResult):
            return self.task_metadata.id == other.task_metadata.id
        return NotImplemented

    def get_compiled_circuit(self) -> Optional[str]:
        """Get the compiled circuit, if one is available.

        Returns:
            Optional[str]: The compiled circuit or None.
        """
        metadata = self.additional_metadata
        if not metadata:
            return None
        if metadata.rigettiMetadata:
            return metadata.rigettiMetadata.compiledProgram
        elif metadata.oqcMetadata:
            return metadata.oqcMetadata.compiledProgram
        else:
            return None

    @staticmethod
    def measurement_counts_from_measurements(measurements: np.ndarray) -> Counter:
        """Creates measurement counts from measurements

        Args:
            measurements (np.ndarray): 2d array - row is shot and column is qubit.

        Returns:
            Counter: A Counter of measurements. Key is the measurements in a big endian binary
            string. Value is the number of times that measurement occurred.
        """
        bitstrings = []
        for j in range(len(measurements)):
            bitstrings.append("".join([str(element) for element in measurements[j]]))
        return Counter(bitstrings)

    @staticmethod
    def measurement_probabilities_from_measurement_counts(
        measurement_counts: Counter,
    ) -> dict[str, float]:
        """Creates measurement probabilities from measurement counts

        Args:
            measurement_counts (Counter): A Counter of measurements. Key is the measurements
                in a big endian binary string. Value is the number of times that measurement
                occurred.

        Returns:
            dict[str, float]: A dictionary of probabilistic results. Key is the measurements
            in a big endian binary string. Value is the probability the measurement occurred.
        """
        measurement_probabilities = {}
        shots = sum(measurement_counts.values())

        for key, count in measurement_counts.items():
            measurement_probabilities[key] = count / shots
        return measurement_probabilities

    @staticmethod
    def measurements_from_measurement_probabilities(
        measurement_probabilities: dict[str, float], shots: int
    ) -> np.ndarray:
        """Creates measurements from measurement probabilities.

        Args:
            measurement_probabilities (dict[str, float]): A dictionary of probabilistic results.
                Key is the measurements in a big endian binary string.
                Value is the probability the measurement occurred.
            shots (int): number of iterations on device.

        Returns:
            np.ndarray: A dictionary of probabilistic results.
            Key is the measurements in a big endian binary string.
            Value is the probability the measurement occurred.
        """
        measurements_list = []
        for bitstring in measurement_probabilities:
            measurement = list(bitstring)
            individual_measurement_list = [measurement] * int(
                round(measurement_probabilities[bitstring] * shots)
            )
            measurements_list.extend(individual_measurement_list)
        return np.asarray(measurements_list, dtype=int)

    @staticmethod
    def from_object(result: GateModelTaskResult) -> GateModelQuantumTaskResult:
        """Create GateModelQuantumTaskResult from GateModelTaskResult object.

        Args:
            result (GateModelTaskResult): GateModelTaskResult object

        Returns:
            GateModelQuantumTaskResult: A GateModelQuantumTaskResult based on the given dict

        Raises:
            ValueError: If neither "Measurements" nor "MeasurementProbabilities" is a key
                in the result dict
        """
        return GateModelQuantumTaskResult._from_object_internal(result)

    @staticmethod
    def from_string(result: str) -> GateModelQuantumTaskResult:
        """Create GateModelQuantumTaskResult from string.

        Args:
            result (str): JSON object string, with GateModelQuantumTaskResult attributes as keys.

        Returns:
            GateModelQuantumTaskResult: A GateModelQuantumTaskResult based on the given string

        Raises:
            ValueError: If neither "Measurements" nor "MeasurementProbabilities" is a key
                in the result dict
        """
        obj = GateModelTaskResult.parse_raw(result)
        GateModelQuantumTaskResult.cast_result_types(obj)
        return GateModelQuantumTaskResult._from_object_internal(obj)

    @classmethod
    def _from_object_internal(cls, result: GateModelTaskResult) -> GateModelQuantumTaskResult:
        if result.taskMetadata.shots > 0:
            return GateModelQuantumTaskResult._from_object_internal_computational_basis_sampling(
                result
            )
        else:
            return GateModelQuantumTaskResult._from_dict_internal_simulator_only(result)

    @classmethod
    def _from_object_internal_computational_basis_sampling(
        cls, result: GateModelTaskResult
    ) -> GateModelQuantumTaskResult:
        task_metadata = result.taskMetadata
        additional_metadata = result.additionalMetadata
        if result.measurements:
            measurements = np.asarray(result.measurements, dtype=int)
            m_counts = GateModelQuantumTaskResult.measurement_counts_from_measurements(measurements)
            m_probs = GateModelQuantumTaskResult.measurement_probabilities_from_measurement_counts(
                m_counts
            )
            measurements_copied_from_device = True
            m_counts_copied_from_device = False
            m_probabilities_copied_from_device = False
        elif result.measurementProbabilities:
            shots = task_metadata.shots
            m_probs = result.measurementProbabilities
            measurements = GateModelQuantumTaskResult.measurements_from_measurement_probabilities(
                m_probs, shots
            )
            m_counts = GateModelQuantumTaskResult.measurement_counts_from_measurements(measurements)
            measurements_copied_from_device = False
            m_counts_copied_from_device = False
            m_probabilities_copied_from_device = True
        else:
            raise ValueError(
                'One of "measurements" or "measurementProbabilities" must be populated in',
                " the result obj",
            )
        measured_qubits = result.measuredQubits
        if result.resultTypes:
            # Jaqcd does not return anything in the resultTypes schema field since the
            # result types are easily parsable from the IR. However, an OpenQASM program
            # specifies result types inline and parsing result types is more involved
            # (ie. may involve dereferencing logical qubits at runtime), so the parsed
            # result type specifications need to be returned, even if not calculated
            # during simulation.
            if not isinstance(result.resultTypes[0], ResultTypeValue):
                result_types = GateModelQuantumTaskResult._calculate_result_types(
                    json.dumps({"results": [rt.dict() for rt in result.resultTypes]}),
                    measurements,
                    measured_qubits,
                )
            else:
                result_types = result.resultTypes
        else:
            result_types = GateModelQuantumTaskResult._calculate_result_types(
                additional_metadata.action.json(), measurements, measured_qubits
            )
        values = [rt.value for rt in result_types]
        return cls(
            task_metadata=task_metadata,
            additional_metadata=additional_metadata,
            result_types=result_types,
            values=values,
            measurements=measurements,
            measured_qubits=measured_qubits,
            measurement_counts=m_counts,
            measurement_probabilities=m_probs,
            measurements_copied_from_device=measurements_copied_from_device,
            measurement_counts_copied_from_device=m_counts_copied_from_device,
            measurement_probabilities_copied_from_device=m_probabilities_copied_from_device,
        )

    @classmethod
    def _from_dict_internal_simulator_only(
        cls, result: GateModelTaskResult
    ) -> GateModelQuantumTaskResult:
        task_metadata = result.taskMetadata
        additional_metadata = result.additionalMetadata
        result_types = result.resultTypes
        values = [rt.value for rt in result_types]
        return cls(
            task_metadata=task_metadata,
            additional_metadata=additional_metadata,
            result_types=result_types,
            values=values,
        )

    @staticmethod
    def cast_result_types(gate_model_task_result: GateModelTaskResult) -> None:
        """Casts the result types to the types expected by the SDK.

        Args:
            gate_model_task_result (GateModelTaskResult): GateModelTaskResult representing the
                results.
        """
        if gate_model_task_result.resultTypes:
            for result_type in gate_model_task_result.resultTypes:
                type = result_type.type.type
                if type == "probability":
                    result_type.value = np.array(result_type.value)
                elif type == "statevector":
                    result_type.value = np.array([complex(*value) for value in result_type.value])
                elif type == "amplitude":
                    for state in result_type.value:
                        result_type.value[state] = complex(*result_type.value[state])

    @staticmethod
    def _calculate_result_types(
        ir_string: str, measurements: np.ndarray, measured_qubits: list[int]
    ) -> list[ResultTypeValue]:
        ir = json.loads(ir_string)
        result_types = []
        if not ir.get("results"):
            return result_types
        for result_type in ir["results"]:
            ir_observable = result_type.get("observable")
            observable = observable_from_ir(ir_observable) if ir_observable else None
            targets = result_type.get("targets")
            rt_type = result_type["type"]
            if rt_type == "probability":
                value = GateModelQuantumTaskResult._probability_from_measurements(
                    measurements, measured_qubits, targets
                )
                casted_result_type = Probability(targets=targets)
            elif rt_type == "sample":
                value = GateModelQuantumTaskResult._calculate_for_targets(
                    GateModelQuantumTaskResult._samples_from_measurements,
                    measurements,
                    measured_qubits,
                    observable,
                    targets,
                )
                casted_result_type = Sample(targets=targets, observable=ir_observable)
            elif rt_type == "variance":
                value = GateModelQuantumTaskResult._calculate_for_targets(
                    GateModelQuantumTaskResult._variance_from_measurements,
                    measurements,
                    measured_qubits,
                    observable,
                    targets,
                )
                casted_result_type = Variance(targets=targets, observable=ir_observable)
            elif rt_type == "expectation":
                value = GateModelQuantumTaskResult._calculate_for_targets(
                    GateModelQuantumTaskResult._expectation_from_measurements,
                    measurements,
                    measured_qubits,
                    observable,
                    targets,
                )
                casted_result_type = Expectation(targets=targets, observable=ir_observable)
            else:
                raise ValueError(f"Unknown result type {rt_type}")
            result_types.append(ResultTypeValue.construct(type=casted_result_type, value=value))
        return result_types

    @staticmethod
    def _selected_measurements(
        measurements: np.ndarray, measured_qubits: list[int], targets: Optional[list[int]]
    ) -> np.ndarray:
        if targets is not None and targets != measured_qubits:
            # Only some qubits targeted
            columns = [measured_qubits.index(t) for t in targets]
            measurements = measurements[:, columns]
        return measurements

    @staticmethod
    def _calculate_for_targets(
        calculate_function: Callable[[np.ndarray, list[int], Observable, list[int]], T],
        measurements: np.ndarray,
        measured_qubits: list[int],
        observable: Observable,
        targets: list[int],
    ) -> Union[T, list[T]]:
        if targets:
            return calculate_function(measurements, measured_qubits, observable, targets)
        else:
            return [
                calculate_function(measurements, measured_qubits, observable, [i])
                for i in measured_qubits
            ]

    @staticmethod
    def _measurements_base_10(measurements: np.ndarray) -> np.ndarray:
        # convert samples from a list of 0, 1 integers, to base 10 representation
        two_powers = 2 ** np.arange(measurements.shape[-1])[::-1]  # 2^(n-1), ..., 2, 1
        return measurements @ two_powers

    @staticmethod
    def _probability_from_measurements(
        measurements: np.ndarray, measured_qubits: list[int], targets: Optional[list[int]]
    ) -> np.ndarray:
        measurements = GateModelQuantumTaskResult._selected_measurements(
            measurements, measured_qubits, targets
        )
        shots, num_measured_qubits = measurements.shape
        # convert measurements from a list of 0, 1 integers, to base 10 representation
        indices = GateModelQuantumTaskResult._measurements_base_10(measurements)

        # count the basis state occurrences, and construct the probability vector
        basis_states, counts = np.unique(indices, return_counts=True)
        probabilities = np.zeros([2**num_measured_qubits], dtype=np.float64)
        probabilities[basis_states] = counts / shots
        return probabilities

    @staticmethod
    def _variance_from_measurements(
        measurements: np.ndarray,
        measured_qubits: list[int],
        observable: Observable,
        targets: list[int],
    ) -> float:
        samples = GateModelQuantumTaskResult._samples_from_measurements(
            measurements, measured_qubits, observable, targets
        )
        return np.var(samples)

    @staticmethod
    def _expectation_from_measurements(
        measurements: np.ndarray,
        measured_qubits: list[int],
        observable: Observable,
        targets: list[int],
    ) -> float:
        samples = GateModelQuantumTaskResult._samples_from_measurements(
            measurements, measured_qubits, observable, targets
        )
        return np.mean(samples)

    @staticmethod
    def _samples_from_measurements(
        measurements: np.ndarray,
        measured_qubits: list[int],
        observable: Observable,
        targets: list[int],
    ) -> np.ndarray:
        measurements = GateModelQuantumTaskResult._selected_measurements(
            measurements, measured_qubits, targets
        )
        if isinstance(observable, StandardObservable):
            # Process samples for observables with eigenvalues {1, -1}
            return 1 - 2 * measurements.flatten()
        # Replace the basis state in the computational basis with the correct eigenvalue.
        # Extract only the columns of the basis samples required based on ``targets``.
        indices = GateModelQuantumTaskResult._measurements_base_10(measurements)
        if isinstance(observable, TensorProduct):
            return np.array([observable.eigenvalue(index).real for index in indices])
        return observable.eigenvalues[indices].real

    @staticmethod
    def _result_type_hash(rt_type: dict) -> str:
        if hasattr(rt_type, "observable") and isinstance(rt_type.observable, list):
            rt_type.observable = GateModelQuantumTaskResult._replace_neg_zero(rt_type.observable)
        return repr(dict(sorted(dict(rt_type).items(), key=lambda x: x[0])))

    @staticmethod
    def _replace_neg_zero(observable_matrix: Union[list, int]) -> Union[list, int]:
        if isinstance(observable_matrix, list):
            return [GateModelQuantumTaskResult._replace_neg_zero(x) for x in observable_matrix]
        else:
            return 0 if observable_matrix == 0 else observable_matrix



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import asyncio
from typing import Union

from braket.tasks import (
    AnnealingQuantumTaskResult,
    GateModelQuantumTaskResult,
    PhotonicModelQuantumTaskResult,
    QuantumTask,
)


class LocalQuantumTask(QuantumTask):
    """A quantum task containing the results of a local simulation.

    Since this class is instantiated with the results, cancel() and run_async() are unsupported.
    """

    def __init__(
        self,
        result: Union[
            GateModelQuantumTaskResult, AnnealingQuantumTaskResult, PhotonicModelQuantumTaskResult
        ],
    ):
        self._id = result.task_metadata.id
        self._result = result

    @property
    def id(self) -> str:
        """Gets the task ID.

        Returns:
            str: The ID of the task.
        """
        return str(self._id)

    def cancel(self) -> None:
        """Cancel the quantum task."""
        raise NotImplementedError("Cannot cancel completed local task")

    def state(self) -> str:
        """Gets the state of the task.

        Returns:
            str: Returns COMPLETED
        """
        return "COMPLETED"

    def result(
        self,
    ) -> Union[
        GateModelQuantumTaskResult, AnnealingQuantumTaskResult, PhotonicModelQuantumTaskResult
    ]:
        return self._result

    def async_result(self) -> asyncio.Task:
        """Get the quantum task result asynchronously.

        Raises:
            NotImplementedError: Asynchronous local simulation unsupported

        Returns:
            asyncio.Task: Get the quantum task result asynchronously.
        """
        # TODO: Allow for asynchronous simulation
        raise NotImplementedError("Asynchronous local simulation unsupported")

    def __repr__(self) -> str:
        return f"LocalQuantumTask('id':{self.id})"



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import json
from unittest.mock import Mock, mock_open, patch

import pytest

from braket.jobs.local.local_job import LocalQuantumJob


@pytest.fixture
def aws_session():
    _aws_session = Mock()
    return _aws_session


@pytest.fixture
def job_results():
    return {"dataFormat": "plaintext", "dataDictionary": {"some_results": {"excellent": "here"}}}


@pytest.fixture
def run_log():
    test_log = (
        "This is a multi-line log.\n"
        "This is the next line.\n"
        "Metrics - timestamp=1633027264.5406773; Cost=-4.034; iteration_number=0;\n"
        "Metrics - timestamp=1633027288.6284382; Cost=-3.957; iteration_number=1;\n"
    )
    return test_log


@pytest.fixture
def test_envs():
    return {"Test": "Env"}


@pytest.mark.parametrize(
    "creation_kwargs",
    [
        (
            {
                "jobName": "Test-Job-Name",
                "algorithmSpecification": {"containerImage": {"uri": "file://test-URI"}},
                "checkpointConfig": {"localPath": "test/local/path/"},
            }
        ),
        (
            {
                "jobName": "Test-Job-Name",
                "algorithmSpecification": {"containerImage": {"uri": "file://test-URI"}},
                "checkpointConfig": {},
            }
        ),
        (
            {
                "jobName": "Test-Job-Name",
                "algorithmSpecification": {"containerImage": {"uri": "file://test-URI"}},
            }
        ),
        (
            {
                "jobName": "Test-Job-Name",
                "algorithmSpecification": {},
            }
        ),
    ],
)
@patch("braket.jobs.local.local_job.prepare_quantum_job")
@patch("braket.jobs.local.local_job.retrieve_image")
@patch("braket.jobs.local.local_job.setup_container")
@patch("braket.jobs.local.local_job._LocalJobContainer")
@patch("os.path.isdir")
def test_create(
    mock_dir,
    mock_container,
    mock_setup,
    mock_retrieve_image,
    mock_prepare_job,
    aws_session,
    creation_kwargs,
    job_results,
    run_log,
    test_envs,
):
    with patch("builtins.open", mock_open()) as file_open:
        mock_dir.return_value = False
        mock_prepare_job.return_value = creation_kwargs

        mock_container_open = mock_container.return_value.__enter__.return_value
        mock_container_open.run_log = run_log
        file_read = file_open()
        file_read.read.return_value = json.dumps(job_results)
        mock_setup.return_value = test_envs

        job = LocalQuantumJob.create(
            device=Mock(),
            source_module=Mock(),
            entry_point=Mock(),
            image_uri=Mock(),
            job_name=Mock(),
            code_location=Mock(),
            role_arn=Mock(),
            hyperparameters=Mock(),
            input_data=Mock(),
            output_data_config=Mock(),
            checkpoint_config=Mock(),
            aws_session=aws_session,
        )
        assert job.name == "Test-Job-Name"
        assert job.arn == "local:job/Test-Job-Name"
        assert job.state() == "COMPLETED"
        assert job.run_log == run_log
        assert job.metadata() is None
        assert job.cancel() is None
        assert job.download_result() is None
        assert job.logs() is None
        assert job.result() == job_results["dataDictionary"]
        assert job.metrics() == {
            "Cost": [-4.034, -3.957],
            "iteration_number": [0.0, 1.0],
            "timestamp": [1633027264.5406773, 1633027288.6284382],
        }
        mock_setup.assert_called_with(mock_container_open, aws_session, **creation_kwargs)
        mock_container_open.run_local_job.assert_called_with(test_envs)


def test_create_invalid_arg():
    unexpected_kwarg = "create\\(\\) got an unexpected keyword argument 'wait_until_complete'"
    with pytest.raises(TypeError, match=unexpected_kwarg):
        LocalQuantumJob.create(
            device="device",
            source_module="source",
            wait_until_complete=True,
        )


@patch("os.path.isdir")
def test_read_runlog_file(mock_dir):
    mock_dir.return_value = True
    with patch("builtins.open", mock_open()) as file_open:
        file_read = file_open()
        file_read.read.return_value = "Test Log"
        job = LocalQuantumJob("local:job/Fake-Job")
        assert job.run_log == "Test Log"


@patch("braket.jobs.local.local_job.prepare_quantum_job")
@patch("os.path.isdir")
def test_create_existing_job(mock_dir, mock_prepare_job, aws_session):
    mock_dir.return_value = True
    mock_prepare_job.return_value = {
        "jobName": "Test-Job-Name",
        "algorithmSpecification": {"containerImage": {"uri": "file://test-URI"}},
        "checkpointConfig": {"localPath": "test/local/path/"},
    }
    dir_already_exists = (
        "A local directory called Test-Job-Name already exists. Please use a different job name."
    )
    with pytest.raises(ValueError, match=dir_already_exists):
        LocalQuantumJob.create(
            device=Mock(),
            source_module=Mock(),
            entry_point=Mock(),
            image_uri=Mock(),
            job_name=Mock(),
            code_location=Mock(),
            role_arn=Mock(),
            hyperparameters=Mock(),
            input_data=Mock(),
            output_data_config=Mock(),
            checkpoint_config=Mock(),
            aws_session=aws_session,
        )


def test_invalid_arn():
    invalid_arn = "Arn Invalid-Arn is not a valid local job arn"
    with pytest.raises(ValueError, match=invalid_arn):
        LocalQuantumJob("Invalid-Arn")


def test_missing_job_dir():
    missing_dir = "Unable to find local job results for Missing-Dir"
    with pytest.raises(ValueError, match=missing_dir):
        LocalQuantumJob("local:job/Missing-Dir")


@patch("os.path.isdir")
def test_missing_runlog_file(mock_dir):
    mock_dir.return_value = True
    job = LocalQuantumJob("local:job/Fake-Dir")
    no_file = "Unable to find logs in the local job directory Fake-Dir."
    with pytest.raises(ValueError, match=no_file):
        job.run_log


@patch("os.path.isdir")
def test_missing_results_file(mock_dir):
    mock_dir.return_value = True
    job = LocalQuantumJob("local:job/Fake-Dir")
    no_results = "Unable to find results in the local job directory Fake-Dir."
    with pytest.raises(ValueError, match=no_results):
        job.result()



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from abc import ABC, abstractmethod
from typing import Union

from braket.tasks.annealing_quantum_task_result import AnnealingQuantumTaskResult
from braket.tasks.gate_model_quantum_task_result import GateModelQuantumTaskResult
from braket.tasks.photonic_model_quantum_task_result import PhotonicModelQuantumTaskResult


class QuantumTaskBatch(ABC):
    """An abstraction over a quantum task batch on a quantum device."""

    @abstractmethod
    def results(
        self,
    ) -> list[
        Union[
            GateModelQuantumTaskResult, AnnealingQuantumTaskResult, PhotonicModelQuantumTaskResult
        ]
    ]:
        """Get the quantum task results.

        Returns:
            list[Union[GateModelQuantumTaskResult, AnnealingQuantumTaskResult, PhotonicModelQuantumTaskResult]]: Get
            the quantum task results.
        """  # noqa: E501



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import annotations


class TrackingContext:
    def __init__(self):
        self._trackers = set()

    def register_tracker(self, tracker: Tracker) -> None:  # noqa F821
        """Registers a tracker.

        Args:
            tracker (Tracker): The tracker.
        """
        self._trackers.add(tracker)

    def deregister_tracker(self, tracker: Tracker) -> None:  # noqa F821
        """Deregisters a tracker.

        Args:
            tracker (Tracker): The tracker.
        """
        self._trackers.remove(tracker)

    def broadcast_event(self, event: _TrackingEvent) -> None:  # noqa F821
        """Broadcasts an event to all trackers.

        Args:
            event (_TrackingEvent): The event to broadcast.
        """
        for tracker in self._trackers:
            tracker.receive_event(event)

    def active_trackers(self) -> set:
        """Gets the active trackers.

        Returns:
            set: The set of active trackers.
        """
        return self._trackers


_tracking_context = TrackingContext()
register_tracker = _tracking_context.register_tracker
deregister_tracker = _tracking_context.deregister_tracker
broadcast_event = _tracking_context.broadcast_event
active_trackers = _tracking_context.active_trackers
